# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""edit entity attributes/relations from any view, without going to the entity
form
"""


import copy

from logilab.common.decorators import cached
from logilab.mtconverter import xml_escape

from cubicweb import _
from cubicweb import neg_role
from cubicweb.predicates import non_final_entity, match_kwargs
from cubicweb.schema import display_name
from cubicweb.utils import json, json_dumps
from cubicweb_web.view import EntityView
from cubicweb_web import stdmsgs
from cubicweb_web.formwidgets import Button, SubmitButton
from cubicweb_web.views.ajaxcontroller import ajaxfunc


class _DummyForm(object):
    __slots__ = ("event_args",)

    def form_render(self, **_args):
        return ""

    def render(self, *_args, **_kwargs):
        return ""

    def append_field(self, *args):
        pass

    def add_hidden(self, *args):
        pass


class AutoClickAndEditFormView(EntityView):
    __regid__ = "reledit"
    __select__ = non_final_entity() & match_kwargs("rtype")

    # ui side continuations
    _onclick = (
        "cw.reledit.loadInlineEditionForm('%(formid)s', %(eid)s, '%(rtype)s', '%(role)s', "
        "'%(divid)s', %(reload)s, '%(vid)s', '%(action)s');"
    )
    _cancelclick = "cw.reledit.cleanupAfterCancel('%s')"

    # ui side actions/buttons
    _addzone = '<img title="%(msg)s" src="%(logo)s" alt="%(msg)s"/>'
    _addmsg = _("click to add a value")
    _addlogo = "plus.png"
    _deletezone = '<img title="%(msg)s" src="%(logo)s" alt="%(msg)s"/>'
    _deletemsg = _("click to delete this value")
    _deletelogo = "cancel.png"
    _editzone = '<img title="%(msg)s" src="%(logo)s" alt="%(msg)s"/>'
    _editzonemsg = _("click to edit this field")
    _editlogo = "pen_icon.png"

    # renderer
    _form_renderer_id = "base"

    def entity_call(
        self,
        entity,
        rtype=None,
        role="subject",
        reload=False,  # controls reloading the whole page after change
        # boolean, eid (to redirect), or
        # function taking the subject entity & returning a boolean or an eid
        rvid=None,  # vid to be applied to other side of rtype (non final relations only)
        default_value=None,
        formid="base",
        action=None,
    ):
        """display field to edit entity's `rtype` relation on click"""
        assert rtype
        self._cw.add_css("cubicweb.form.css")
        self._cw.add_js(
            ("cubicweb.reledit.js", "cubicweb.edition.js", "cubicweb.ajax.js")
        )
        self.entity = entity
        rschema = self._cw.vreg.schema[rtype]
        rctrl = self._cw.vreg["uicfg"].select("reledit", self._cw, entity=entity)
        self._rules = rctrl.etype_get(
            self.entity.e_schema.type, rschema.type, role, "*"
        )
        reload = self._compute_reload(rschema, role, reload)
        divid = self._build_divid(rtype, role, self.entity.eid)
        if rschema.final:
            self._handle_attribute(rschema, role, divid, reload, action)
        else:
            if self._is_composite():
                self._handle_composite(rschema, role, divid, reload, formid, action)
            else:
                self._handle_relation(rschema, role, divid, reload, formid, action)

    def _handle_attribute(self, rschema, role, divid, reload, action):
        rvid = self._rules.get("rvid", None)
        if rvid is not None:
            value = self._cw.view(
                rvid, entity=self.entity, rtype=rschema.type, role=role
            )
        else:
            value = self.entity.printable_value(rschema.type)
        if not self._should_edit_attribute(rschema):
            self.w(value)
            return
        form, renderer = self._build_form(
            self.entity, rschema, role, divid, "base", reload, action
        )
        value = value or self._compute_default_value(rschema, role)
        self.view_form(divid, value, form, renderer)

    def _compute_formid_value(self, rschema, role, rvid, formid):
        related_rset = self.entity.related(rschema.type, role)
        if related_rset:
            value = self._cw.view(rvid, related_rset)
        else:
            value = self._compute_default_value(rschema, role)
        if not self._should_edit_relation(rschema, role):
            return None, value
        return formid, value

    def _handle_relation(self, rschema, role, divid, reload, formid, action):
        rvid = self._rules.get("rvid", "autolimited")
        formid, value = self._compute_formid_value(rschema, role, rvid, formid)
        if formid is None:
            return self.w(value)
        form, renderer = self._build_form(
            self.entity, rschema, role, divid, formid, reload, action, dict(vid=rvid)
        )
        self.view_form(divid, value, form, renderer)

    def _handle_composite(self, rschema, role, divid, reload, formid, action):
        # this is for attribute-like composites (1 target type, 1 related entity at most, for now)
        entity = self.entity
        related_rset = entity.related(rschema.type, role)
        add_related = self._may_add_related(related_rset, rschema, role)
        edit_related = self._may_edit_related_entity(related_rset, rschema, role)
        delete_related = edit_related and self._may_delete_related(
            related_rset, rschema, role
        )
        rvid = self._rules.get("rvid", "autolimited")
        formid, value = self._compute_formid_value(rschema, role, rvid, formid)
        if formid is None or not (edit_related or add_related):
            # till we learn to handle cases where not (edit_related or add_related)
            self.w(value)
            return
        form, renderer = self._build_form(
            entity, rschema, role, divid, formid, reload, action, dict(vid=rvid)
        )
        self.view_form(
            divid, value, form, renderer, edit_related, add_related, delete_related
        )

    @cached
    def _compute_ttypes(self, rschema, role):
        dual_role = neg_role(role)
        return getattr(rschema, "%ss" % dual_role)(self.entity.e_schema)

    def _compute_reload(self, rschema, role, reload):
        ctrl_reload = self._rules.get("reload", reload)
        if callable(ctrl_reload):
            ctrl_reload = ctrl_reload(self.entity)
        if isinstance(ctrl_reload, int) and ctrl_reload > 1:  # not True/False
            ctrl_reload = self._cw.build_url(ctrl_reload)
        return ctrl_reload

    def _compute_default_value(self, rschema, role):
        default = self._rules.get("novalue_label")
        if default is None:
            if self._rules.get("novalue_include_rtype"):
                default = self._cw._("<%s not specified>") % display_name(
                    self._cw, rschema.type, role
                )
            else:
                default = self._cw._("<not specified>")
        else:
            default = self._cw._(default)
        return xml_escape(default)

    def _is_composite(self):
        return self._rules.get("edit_target") == "related"

    def _may_add_related(self, related_rset, rschema, role):
        """ok for attribute-like composite entities"""
        ttypes = self._compute_ttypes(rschema, role)
        if len(ttypes) > 1:  # many etypes: learn how to do it
            return False
        rdef = rschema.role_relation_definition(self.entity.e_schema, ttypes[0], role)
        card = rdef.role_cardinality(role)
        if related_rset or card not in "?1":
            return False
        if role == "subject":
            kwargs = {"fromeid": self.entity.eid}
        else:
            kwargs = {"toeid": self.entity.eid}
        return rdef.has_perm(self._cw, "add", **kwargs)

    def _may_edit_related_entity(self, related_rset, rschema, role):
        """controls the edition of the related entity"""
        ttypes = self._compute_ttypes(rschema, role)
        if len(ttypes) > 1 or len(related_rset.rows) != 1:
            return False
        if (
            self.entity.e_schema.relation_definition(rschema, role).role_cardinality(
                role
            )
            not in "?1"
        ):
            return False
        return related_rset.get_entity(0, 0).cw_has_perm("update")

    def _may_delete_related(self, related_rset, rschema, role):
        # we assume may_edit_related, only 1 related entity
        if not related_rset:
            return False
        rentity = related_rset.get_entity(0, 0)
        entity = self.entity
        if role == "subject":
            kwargs = {"fromeid": entity.eid, "toeid": rentity.eid}
            cardinality = rschema.relation_definitions[
                (entity.cw_etype, rentity.cw_etype)
            ].cardinality[0]
        else:
            kwargs = {"fromeid": rentity.eid, "toeid": entity.eid}
            cardinality = rschema.rdefs[
                (rentity.cw_etype, entity.cw_etype)
            ].cardinality[1]
        if cardinality in "1+":
            return False
        # NOTE: should be sufficient given a well built schema/security
        return rschema.has_perm(self._cw, "delete", **kwargs)

    def _build_zone(self, zonedef, msg, logo):
        return zonedef % {
            "msg": xml_escape(self._cw._(msg)),
            "logo": xml_escape(self._cw.data_url(logo)),
        }

    def _build_edit_zone(self):
        return self._build_zone(self._editzone, self._editzonemsg, self._editlogo)

    def _build_delete_zone(self):
        return self._build_zone(self._deletezone, self._deletemsg, self._deletelogo)

    def _build_add_zone(self):
        return self._build_zone(self._addzone, self._addmsg, self._addlogo)

    def _build_divid(self, rtype, role, entity_eid):
        """builds an id for the root div of a reledit widget"""
        return "%s-%s-%s" % (rtype, role, entity_eid)

    def _build_args(self, entity, rtype, role, formid, reload, action, extradata=None):
        divid = self._build_divid(rtype, role, entity.eid)
        event_args = {
            "divid": divid,
            "eid": entity.eid,
            "rtype": rtype,
            "formid": formid,
            "reload": json_dumps(reload),
            "action": action,
            "role": role,
            "vid": "",
        }
        if extradata:
            event_args.update(extradata)
        return event_args

    def _prepare_form(self, entity, rschema, role, action):
        assert action in ("edit_rtype", "edit_related", "add", "delete"), action
        if action == "edit_rtype":
            return False, entity
        label = True
        if action in ("edit_related", "delete"):
            edit_entity = entity.related(rschema, role).get_entity(0, 0)
        elif action == "add":
            add_etype = self._compute_ttypes(rschema, role)[0]
            _new_entity = self._cw.vreg["etypes"].etype_class(add_etype)(self._cw)
            _new_entity.eid = next(self._cw.varmaker)
            edit_entity = _new_entity
            # XXX see forms.py ~ 276 and entities.linked_to method
            #     is there another way?
            self._cw.form["__linkto"] = "%s:%s:%s" % (
                rschema,
                entity.eid,
                neg_role(role),
            )
        assert edit_entity
        return label, edit_entity

    def _build_renderer(self, related_entity, display_label):
        return self._cw.vreg["formrenderers"].select(
            self._form_renderer_id,
            self._cw,
            entity=related_entity,
            display_label=display_label,
            table_class="attributeForm" if display_label else "",
            display_help=False,
            button_bar_class="buttonbar",
            display_progress_div=False,
        )

    def _build_form(
        self,
        entity,
        rschema,
        role,
        divid,
        formid,
        reload,
        action,
        extradata=None,
        **formargs,
    ):
        rtype = rschema.type
        event_args = self._build_args(
            entity, rtype, role, formid, reload, action, extradata
        )
        if not action:
            form = _DummyForm()
            form.event_args = event_args
            return form, None
        label, edit_entity = self._prepare_form(entity, rschema, role, action)
        cancelclick = self._cancelclick % divid
        form = self._cw.vreg["forms"].select(
            formid,
            self._cw,
            rset=edit_entity.as_rset(),
            entity=edit_entity,
            domid="%s-form" % divid,
            formtype="inlined",
            action=self._cw.build_url(
                "validateform", __onsuccess="window.parent.cw.reledit.onSuccess"
            ),
            cwtarget="eformframe",
            cssclass="releditForm",
            **formargs,
        )
        # pass reledit arguments
        for pname, pvalue in event_args.items():
            form.add_hidden("__reledit|" + pname, pvalue)
        # handle buttons
        if form.form_buttons:  # edition, delete
            form_buttons = []
            for button in form.form_buttons:
                if not button.label.endswith("apply"):
                    if button.label.endswith("cancel"):
                        button = copy.deepcopy(button)
                        button.cwaction = None
                        button.onclick = cancelclick
                        if "class" in button.attrs:
                            new_class = button.attrs["class"].replace(
                                "cwjs-edition-cancel", ""
                            )
                            button.attrs["class"] = new_class
                    form_buttons.append(button)
            form.form_buttons = form_buttons
        else:  # base
            form.form_buttons = [
                SubmitButton(),
                Button(stdmsgs.BUTTON_CANCEL, onclick=cancelclick),
            ]
        form.event_args = event_args
        if formid == "base":
            field = form.field_by_name(rtype, role, entity.e_schema)
            form.append_field(field)
        return form, self._build_renderer(edit_entity, label)

    def _should_edit_attribute(self, rschema):
        entity = self.entity
        rdef = entity.e_schema.relation_definition(rschema)
        # check permissions
        if not entity.cw_has_perm("update"):
            return False
        rdef = entity.e_schema.relation_definition(rschema)
        return rdef.has_perm(self._cw, "update", eid=entity.eid)

    def _may_delete_related_entities(self, related_rset, rschema, ttypes, role):
        rdef = rschema.role_relation_definition(self.entity.e_schema, ttypes[0], role)
        if role == "subject":
            # on relation with cardinality 1 or ?, we need delete perm as well
            # if the relation is already set
            if rdef.role_cardinality(role) in "1?" and not any(
                rdef.has_perm(self._cw, "delete", fromeid=self.entity.eid, toeid=row[0])
                for row in related_rset
            ):
                return False
        elif role == "object":
            # on relation with cardinality 1 or ?, we need delete perm as well
            # if the relation is already set
            if rdef.role_cardinality(role) in "1?" and not any(
                rdef.has_perm(self._cw, "delete", toeid=self.entity.eid, fromeid=row[0])
                for row in related_rset
            ):
                return False
        return True

    def _has_unrelated_entities_to_add(self, rschema, ttypes, role, **perm_args):
        """can add at least one unrelated entity"""
        for target in ttypes:
            if self.entity.unrelated(rschema, target, role, limit=1):
                return True
        return False

    def _should_edit_relation(self, rschema, role):
        """reledit for an relation if one of following conditions is satisfied:
        * the relation dont have rqlexpr permissions and can be deleted
        * at least one of related entites can be deleted

        These two conditions are consistent with thoses in
        AutoformSectionRelationTags.relations_by_section.

        We also check that at least one unrelated entity exists and can be added

        """

        eeid = self.entity.eid
        perm_args = {"fromeid": eeid} if role == "subject" else {"toeid": eeid}
        ttypes = self._compute_ttypes(rschema, role)
        # check permission which are not rqlexprs
        for target in ttypes:
            rdef = rschema.role_relation_definition(self.entity.e_schema, target, role)
            if rdef.has_perm(self._cw, "add", **perm_args):
                return True
        # can delete related entities
        related_rset = self.entity.related(rschema.type, role)
        if related_rset:
            if not self._may_delete_related_entities(
                related_rset, rschema, ttypes, role
            ):
                return False
            return True
        # there are unrelated entities to be added
        if self._has_unrelated_entities_to_add(rschema, ttypes, role, **perm_args):
            return True
        # no related or unrelated entities exist
        # consider autoform_permissions_overrides
        permsoverrides = self._cw.vreg["uicfg"].select(
            "autoform_permissions_overrides", self._cw, entity=self.entity
        )
        if "add_on_new" in permsoverrides.etype_get(
            self.entity.e_schema, rschema, role
        ):
            return True
        # there no way to verify subject/object related permissions as no unrelated
        # entities exist. In doubt we dont display the reledit for security reasons
        return False

    def _open_form_wrapper(
        self, divid, value, form, renderer, _edit_related, _add_related, _delete_related
    ):
        self.w(
            '<div id="%(id)s-reledit" onmouseout="%(out)s" onmouseover="%(over)s" class="%(css)s">',
            {
                "id": divid,
                "css": "releditField",
                "out": "jQuery('#%s').addClass('invisible')" % divid,
                "over": "jQuery('#%s').removeClass('invisible')" % divid,
            },
            escape=False,
        )
        self.w('<div id="%s-value" class="editableFieldValue">', divid)
        self.w(value)
        self.w("</div>")
        form.render(w=self.w, renderer=renderer)
        self.w('<div id="%s" class="editableField invisible">', divid)

    def _edit_action(self, divid, args, edit_related, add_related, _delete_related):
        # XXX disambiguate wrt edit_related
        if not add_related:  # currently, excludes edition
            args["formid"] = "edition" if edit_related else "base"
            args["action"] = "edit_related" if edit_related else "edit_rtype"
            self.w(
                '<div id="%s-update" class="editableField" onclick="%s" title="%s">',
                divid,
                self._onclick % args,
                self._cw._(self._editzonemsg),
            )
            self.w(self._build_edit_zone())
            self.w("</div>")

    def _add_action(self, divid, args, _edit_related, add_related, _delete_related):
        if add_related:
            args["formid"] = "edition"
            args["action"] = "add"
            self.w(
                '<div id="%s-add" class="editableField" onclick="%s" title="%s">',
                divid,
                self._onclick % args,
                self._cw._(self._addmsg),
            )
            self.w(self._build_add_zone())
            self.w("</div>")

    def _del_action(self, divid, args, _edit_related, _add_related, delete_related):
        if delete_related:
            args["formid"] = "deleteconf"
            args["action"] = "delete"
            self.w(
                '<div id="%s-delete" class="editableField" onclick="%s" title="%s">',
                divid,
                self._onclick % args,
                self._cw._(self._deletemsg),
            )
            self.w(self._build_delete_zone())
            self.w("</div>")

    def _close_form_wrapper(self):
        self.w("</div>")
        self.w("</div>")

    def view_form(
        self,
        divid,
        value,
        form=None,
        renderer=None,
        edit_related=False,
        add_related=False,
        delete_related=False,
    ):
        self._open_form_wrapper(
            divid, value, form, renderer, edit_related, add_related, delete_related
        )
        args = form.event_args.copy()
        self._edit_action(divid, args, edit_related, add_related, delete_related)
        self._add_action(divid, args, edit_related, add_related, delete_related)
        self._del_action(divid, args, edit_related, add_related, delete_related)
        self._close_form_wrapper()


@ajaxfunc(output_type="xhtml")
def reledit_form(self):
    req = self._cw
    args = dict(
        (x, req.form[x]) for x in ("formid", "rtype", "role", "reload", "action")
    )
    rset = req.eid_rset(int(self._cw.form["eid"]))
    try:
        args["reload"] = json.loads(args["reload"])
    except ValueError:  # not true/false, an absolute url
        assert args["reload"].startswith("http")
    view = req.vreg["views"].select("reledit", req, rset=rset, rtype=args["rtype"])
    return self._call_view(view, **args)



# copyright 2019-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

import inspect

from cubicweb_web.devtools.testlib import WebCWTC
from cubicweb_web.view import View


class DebugHtmlRenderingCWTC(WebCWTC):
    def setUp(self):
        super().setUp()
        View.debug_html_rendering = True

    def tearDown(self):
        super().tearDown()
        View.debug_html_rendering = False

    def test_debug_html_rendering_inject_tags(self):
        with self.admin_access.web_request() as req:
            view = self.vreg["views"].select("index", req)
            view_class = view.__class__
            page = view.render()

            self.assertIn(
                'cubicweb-generated-by="%s.%s"'
                % (view_class.__module__, view_class.__name__),
                page,
            )
            source_file = inspect.getsourcefile(view_class)
            self.assertIn('cubicweb-from-source="%s' % (source_file), page)


if __name__ == "__main__":
    from logilab.common.testlib import unittest_main

    unittest_main()



from urllib.parse import urlparse, urljoin

from cubicweb.devtools import BASE_URL
from cubicweb_web.devtools.testlib import PyramidWebCWTC


def set_language(request):
    lang = request.POST.get("lang", None)
    cnx = request.cw_cnx
    if lang is None:
        cnx.execute(
            "DELETE CWProperty X WHERE X for_user U, U eid %(u)s", {"u": cnx.user.eid}
        )
    else:
        cnx.user.set_property("ui.language", lang)
    cnx.commit()

    request.response.text = cnx.user.properties.get("ui.language", "")
    return request.response


def add_remove_group(request):
    add_remove = request.POST["add_remove"]
    cnx = request.cw_cnx
    if add_remove == "add":
        cnx.execute(
            'SET U in_group G WHERE G name "users", U eid %(u)s', {"u": cnx.user.eid}
        )
    else:
        cnx.execute(
            'DELETE U in_group G WHERE G name "users", U eid %(u)s', {"u": cnx.user.eid}
        )
    cnx.commit()

    request.response.text = ",".join(sorted(cnx.user.groups))
    return request.response


class SessionSyncHoooksTC(PyramidWebCWTC):
    def includeme(self, config):
        config.route_prefix = urlparse(BASE_URL).path
        for view in (set_language, add_remove_group):
            config.add_route(view.__name__, "/" + view.__name__)
            config.add_view(view, route_name=view.__name__)

        super().includeme(config)

    def setUp(self):
        super().setUp()
        with self.admin_access.repo_cnx() as cnx:
            self.admin_eid = cnx.user.eid

    def test_sync_props(self):
        # grab the crsf token before login
        response = self.webapp.get(urljoin(BASE_URL, "login"))
        csrf_token = response.forms[0].fields["csrf_token"][0].value

        # initialize a pyramid session using admin credentials
        res = self.webapp.post(
            urljoin(BASE_URL, "login"),
            {"__login": self.admlogin, "__password": self.admpassword},
        )
        self.assertEqual(res.status_int, 303)
        # new property
        res = self.webapp.post(
            urljoin(BASE_URL, "set_language"), {"lang": "fr", "csrf_token": csrf_token}
        )
        self.assertEqual(res.text, "fr")
        # updated property
        res = self.webapp.post(
            urljoin(BASE_URL, "set_language"), {"lang": "en", "csrf_token": csrf_token}
        )
        self.assertEqual(res.text, "en")
        # removed property
        res = self.webapp.post(
            urljoin(BASE_URL, "set_language"), {"csrf_token": csrf_token}
        )
        self.assertEqual(res.text, "")

    def test_sync_groups(self):
        self.webapp.login(self.admlogin, self.admpassword)
        # XXX how to get pyramid request using this session?
        res = self.webapp.post(
            urljoin(BASE_URL, "add_remove_group"), {"add_remove": "add"}
        )
        self.assertEqual(res.text, "managers,users")
        res = self.webapp.post(
            urljoin(BASE_URL, "add_remove_group"), {"add_remove": "remove"}
        )
        self.assertEqual(res.text, "managers")


if __name__ == "__main__":
    from unittest import main

    main()



# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
import re
import sys

from logilab.common.testlib import unittest_main

from cubicweb import Forbidden

from cubicweb_web.devtools.testlib import WebCWTC
from cubicweb_web.view import StartupView
from cubicweb_web import Redirect


class ErrorViewCWTC(WebCWTC):
    def setUp(self):
        super().setUp()
        self.vreg.config["submit-mail"] = "test@logilab.fr"
        self.vreg.config["print-traceback"] = "yes"

    def test_error_generation(self):
        """
        tests
        """

        class MyWrongView(StartupView):
            __regid__ = "my-view"

            def call(self):
                raise ValueError("This is wrong")

        with self.temporary_appobjects(MyWrongView):
            with self.admin_access.web_request() as req:
                try:
                    self.view("my-view", req=req)
                except Exception as e:
                    req.data["excinfo"] = sys.exc_info()
                    req.data["ex"] = e
                    html = self.view("error", req=req)
                    self.assertTrue(
                        re.search(
                            b'^<input name="__signature" type="hidden" '
                            b'value="[0-9a-f]{128}" />$',
                            html.source,
                            re.M,
                        )
                    )

    def test_error_submit_nosig(self):
        """
        tests that the reportbug controller refuses submission if
        there is not content signature
        """
        with self.admin_access.web_request() as req:
            req.form = {"description": "toto"}
            with self.assertRaises(Forbidden):
                self.ctrl_publish(req, "reportbug")

    def test_error_submit_wrongsig(self):
        """
        tests that the reportbug controller refuses submission if the
        content signature is invalid
        """
        with self.admin_access.web_request() as req:
            req.form = {"__signature": "X", "description": "toto"}
            with self.assertRaises(Forbidden):
                self.ctrl_publish(req, "reportbug")

    def test_error_submit_ok(self):
        """
        tests that the reportbug controller accept the email submission if the
        content signature is valid
        """
        with self.admin_access.web_request() as req:
            sign = self.vreg.config.sign_text("toto")
            req.form = {"__signature": sign, "description": "toto"}
            with self.assertRaises(Redirect):
                self.ctrl_publish(req, "reportbug")


if __name__ == "__main__":
    unittest_main()



# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from logilab.common.testlib import TestCase, unittest_main, tag, Tags

from cubicweb_web.devtools.testlib import FakeRequest


def _test_cache(hin, hout, method="GET"):
    """forge and process an HTTP request using given headers in/out and method,
    then return it once its .is_client_cache_valid() method has been called.

    req.status_out is None if the page should have been calculated.
    """
    # forge request
    req = FakeRequest(method=method)
    for key, value in hin:
        req._headers_in.addRawHeader(key, str(value))
    for key, value in hout:
        req.headers_out.addRawHeader(key, str(value))
    # process
    req.status_out = None
    req.is_client_cache_valid()
    return req


class HTTPCache(TestCase):
    """Check that the http cache logic works as expected
    (as far as we understood the RFC)

    """

    tags = TestCase.tags | Tags("http", "cache")

    def assertCache(self, expected, status, situation=""):
        """simple assert for nicer message"""
        if expected != status:
            if expected is None:
                expected = "MODIFIED"
            if status is None:
                status = "MODIFIED"
            msg = f"expected {expected!r} got {status!r}"
            if situation:
                msg = f"{msg} - when: {situation}"
            self.fail(msg)

    def test_IN_none_OUT_none(self):
        #: test that no caching is requested when not data is available
        #: on any side
        req = _test_cache((), ())
        self.assertIsNone(req.status_out)

    def test_IN_Some_OUT_none(self):
        #: test that no caching is requested when no data is available
        #: server (origin) side
        hin = [
            ("if-modified-since", "Sat, 14 Apr 2012 14:39:32 GM"),
        ]
        req = _test_cache(hin, ())
        self.assertIsNone(req.status_out)
        hin = [
            ("if-none-match", "babar/huitre"),
        ]
        req = _test_cache(hin, ())
        self.assertIsNone(req.status_out)
        hin = [
            ("if-modified-since", "Sat, 14 Apr 2012 14:39:32 GM"),
            ("if-none-match", "babar/huitre"),
        ]
        req = _test_cache(hin, ())
        self.assertIsNone(req.status_out)

    def test_IN_none_OUT_Some(self):
        #: test that no caching is requested when no data is provided
        #: by the client
        hout = [
            ("last-modified", "Sat, 14 Apr 2012 14:39:32 GM"),
        ]
        req = _test_cache((), hout)
        self.assertIsNone(req.status_out)
        hout = [
            ("etag", "babar/huitre"),
        ]
        req = _test_cache((), hout)
        self.assertIsNone(req.status_out)
        hout = [
            ("last-modified", "Sat, 14 Apr 2012 14:39:32 GM"),
            ("etag", "babar/huitre"),
        ]
        req = _test_cache((), hout)
        self.assertIsNone(req.status_out)

    @tag("last_modified")
    def test_last_modified_newer(self):
        #: test the proper behavior of modification date only
        # newer
        hin = [
            ("if-modified-since", "Sat, 13 Apr 2012 14:39:32 GM"),
        ]
        hout = [
            ("last-modified", "Sat, 14 Apr 2012 14:39:32 GM"),
        ]
        req = _test_cache(hin, hout)
        self.assertCache(None, req.status_out, "origin is newer than client")

    @tag("last_modified")
    def test_last_modified_older(self):
        # older
        hin = [
            ("if-modified-since", "Sat, 15 Apr 2012 14:39:32 GM"),
        ]
        hout = [
            ("last-modified", "Sat, 14 Apr 2012 14:39:32 GM"),
        ]
        req = _test_cache(hin, hout)
        self.assertCache(304, req.status_out, "origin is older than client")

    @tag("last_modified")
    def test_last_modified_same(self):
        # same
        hin = [
            ("if-modified-since", "Sat, 14 Apr 2012 14:39:32 GM"),
        ]
        hout = [
            ("last-modified", "Sat, 14 Apr 2012 14:39:32 GM"),
        ]
        req = _test_cache(hin, hout)
        self.assertCache(304, req.status_out, "origin is equal to client")

    @tag("etag")
    def test_etag_mismatch(self):
        #: test the proper behavior of etag only
        # etag mismatch
        hin = [
            ("if-none-match", "babar"),
        ]
        hout = [
            ("etag", "celestine"),
        ]
        req = _test_cache(hin, hout)
        self.assertCache(None, req.status_out, "etag mismatch")

    @tag("etag")
    def test_etag_match(self):
        # etag match
        hin = [
            ("if-none-match", "babar"),
        ]
        hout = [
            ("etag", "babar"),
        ]
        req = _test_cache(hin, hout)
        self.assertCache(304, req.status_out, "etag match")
        self.assertEqual(req.headers_out.getRawHeaders("etag"), ["babar"])
        # etag match in multiple
        hin = [
            ("if-none-match", "loutre"),
            ("if-none-match", "babar"),
        ]
        hout = [
            ("etag", "babar"),
        ]
        req = _test_cache(hin, hout)
        self.assertCache(304, req.status_out, "etag match in multiple")
        self.assertEqual(req.headers_out.getRawHeaders("etag"), ["babar"])
        # client use "*" as etag
        hin = [
            ("if-none-match", "*"),
        ]
        hout = [
            ("etag", "babar"),
        ]
        req = _test_cache(hin, hout)
        self.assertCache(304, req.status_out, 'client use "*" as etag')
        self.assertEqual(req.headers_out.getRawHeaders("etag"), ["babar"])

    @tag("etag", "last_modified")
    def test_both(self):
        #: test the proper behavior of etag only
        # both wrong
        hin = [
            ("if-none-match", "babar"),
            ("if-modified-since", "Sat, 14 Apr 2012 14:39:32 GM"),
        ]
        hout = [
            ("etag", "loutre"),
            ("last-modified", "Sat, 15 Apr 2012 14:39:32 GM"),
        ]
        req = _test_cache(hin, hout)
        self.assertCache(None, req.status_out, "both wrong")

    @tag("etag", "last_modified")
    def test_both_etag_mismatch(self):
        # both etag mismatch
        hin = [
            ("if-none-match", "babar"),
            ("if-modified-since", "Sat, 14 Apr 2012 14:39:32 GM"),
        ]
        hout = [
            ("etag", "loutre"),
            ("last-modified", "Sat, 13 Apr 2012 14:39:32 GM"),
        ]
        req = _test_cache(hin, hout)
        self.assertCache(None, req.status_out, "both  but etag mismatch")

    @tag("etag", "last_modified")
    def test_both_but_modified(self):
        # both but modified
        hin = [
            ("if-none-match", "babar"),
            ("if-modified-since", "Sat, 14 Apr 2012 14:39:32 GM"),
        ]
        hout = [
            ("etag", "babar"),
            ("last-modified", "Sat, 15 Apr 2012 14:39:32 GM"),
        ]
        req = _test_cache(hin, hout)
        self.assertCache(None, req.status_out, "both  but modified")

    @tag("etag", "last_modified")
    def test_both_ok(self):
        # both ok
        hin = [
            ("if-none-match", "babar"),
            ("if-modified-since", "Sat, 14 Apr 2012 14:39:32 GM"),
        ]
        hout = [
            ("etag", "babar"),
            ("last-modified", "Sat, 13 Apr 2012 14:39:32 GM"),
        ]
        req = _test_cache(hin, hout)
        self.assertCache(304, req.status_out, "both ok")
        self.assertEqual(req.headers_out.getRawHeaders("etag"), ["babar"])

    @tag("etag", "HEAD")
    def test_head_verb(self):
        #: check than FOUND 200 is properly raise without content on HEAD request
        #: This logic does not really belong here :-/
        # modified
        hin = [
            ("if-none-match", "babar"),
        ]
        hout = [
            ("etag", "rhino/really-not-babar"),
        ]
        req = _test_cache(hin, hout, method="HEAD")
        self.assertCache(None, req.status_out, "modifier HEAD verb")
        # not modified
        hin = [
            ("if-none-match", "babar"),
        ]
        hout = [
            ("etag", "babar"),
        ]
        req = _test_cache(hin, hout, method="HEAD")
        self.assertCache(304, req.status_out, "not modifier HEAD verb")
        self.assertEqual(req.headers_out.getRawHeaders("etag"), ["babar"])

    @tag("etag", "POST")
    def test_post_verb(self):
        # modified
        hin = [
            ("if-none-match", "babar"),
        ]
        hout = [
            ("etag", "rhino/really-not-babar"),
        ]
        req = _test_cache(hin, hout, method="POST")
        self.assertCache(None, req.status_out, "modifier HEAD verb")
        # not modified
        hin = [
            ("if-none-match", "babar"),
        ]
        hout = [
            ("etag", "babar"),
        ]
        req = _test_cache(hin, hout, method="POST")
        self.assertCache(412, req.status_out, "not modifier HEAD verb")


if __name__ == "__main__":
    unittest_main()



# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Set of base controllers, which are directly plugged into the application
object to handle publication.
"""

import http.client
import logging

from cubicweb import (
    NoSelectableObject,
    ObjectNotFound,
    ValidationError,
    AuthenticationError,
    UndoTransactionException,
    Forbidden,
    rdf,
)
from cubicweb._exceptions import NotAnEntity
from cubicweb.predicates import authenticated_user, anonymous_user, match_form_params
from cubicweb.utils import json_dumps
from cubicweb_web import Redirect, NotFound
from cubicweb_web.controller import Controller
from cubicweb_web.views import vid_from_rset

logger = logging.getLogger(__name__)


class LoginController(Controller):
    __regid__ = "login"
    __select__ = anonymous_user()

    def publish(self, rset=None):
        """log in the instance"""
        if self._cw.vreg.config["auth-mode"] == "http":
            # HTTP authentication
            raise AuthenticationError()
        else:
            # Cookie authentication
            self._cw.status_out = http.client.FORBIDDEN
            return self.appli.need_login_content(self._cw)


class LoginControllerForAuthed(Controller):
    __regid__ = "login"
    __select__ = ~anonymous_user()

    def publish(self, rset=None):
        """log in the instance"""
        path = self._cw.form.get("postlogin_path", "")
        # Redirect expects a URL, not a path. Also path may contain a query
        # string, hence should not be given to _cw.build_url()
        raise Redirect(self._cw.base_url() + path)


class LogoutController(Controller):
    __regid__ = "logout"

    def publish(self, rset=None):
        """logout from the instance"""
        return self.appli.session_handler.logout(self._cw, self.goto_url())

    def goto_url(self):
        # * in http auth mode, url will be ignored
        # * in cookie mode redirecting to the index view is enough : either
        #   anonymous connection is allowed and the page will be displayed or
        #   we'll be redirected to the login form
        return self._cw.build_url("view", vid="loggedout")


class ViewController(Controller):
    """standard entry point :
    - build result set
    - select and call main template
    """

    __regid__ = "view"
    template = "main-template"

    def publish(self, rset=None):
        """publish a request, returning an encoded string.

        :raise NotFound: if the rql has been rewritten (rewrite rules) and the
        resulting rset is empty"""
        view, rset = self._select_view_and_rset(rset)
        if (
            not rset
            and getattr(self._cw, "_rql_rewritten", False)
            and self._cw.form.get("empty_rset_raises_404", False)
        ):
            raise NotFound

        if hasattr(self._cw, "path"):
            logger.info(
                f"on {self._cw.path} controller {self} selected view {view} for rset"
                f" {rset}"
            )

        view.set_http_cache_headers()
        if self._cw.is_client_cache_valid():
            return b""

        template = self.appli.main_template_id(self._cw)
        return self._cw.vreg["views"].main_template(
            self._cw, template, rset=rset, view=view
        )

    def _select_view_and_rset(self, rset):
        req = self._cw
        if rset is None and not hasattr(req, "_rql_processed"):
            req._rql_processed = True
            if req.cnx:
                rset = self.process_rql()
            else:
                rset = None
        vid = req.form.get("vid") or vid_from_rset(req, rset, self._cw.vreg.schema)
        if rset and len(rset) == 1:
            for mimetype in rdf.RDF_MIMETYPE_TO_FORMAT:
                try:
                    req.headers_out.addRawHeader(
                        "Link",
                        "<%s>;rel=alternate;type=%s"
                        % (f"{self._cw.base_url()}{rset.one().rest_path()}", mimetype),
                    )
                except NotAnEntity:
                    # was a string, or a number (but not an eid)
                    break
        try:
            view = self._cw.vreg["views"].select(vid, req, rset=rset)
        except ObjectNotFound:
            self.warning("the view %s could not be found", vid)
            req.set_message(req._("The view %s could not be found") % vid)
            vid = vid_from_rset(req, rset, self._cw.vreg.schema)
            view = self._cw.vreg["views"].select(vid, req, rset=rset)
        except NoSelectableObject:
            if rset:
                req.set_message(
                    req._("The view %s can not be applied to this query") % vid
                )
            else:
                req.set_message(
                    req._(
                        "You have no access to this view or it can not "
                        "be used to display the current data."
                    )
                )
            vid = req.form.get("fallbackvid") or vid_from_rset(
                req, rset, req.vreg.schema
            )
            view = req.vreg["views"].select(vid, req, rset=rset)
        return view, rset

    def execute_linkto(self, eid=None):
        """XXX __linkto parameter may cause security issue

        defined here since custom application controller inheriting from this
        one use this method?
        """
        req = self._cw
        if "__linkto" not in req.form:
            return
        if eid is None:
            eid = int(req.form["eid"])
        for linkto in req.list_form_param("__linkto", pop=True):
            rtype, eids, target = linkto.split(":")
            assert target in ("subject", "object")
            eids = eids.split("_")
            if target == "subject":
                rql = "SET X %s Y WHERE X eid %%(x)s, Y eid %%(y)s" % rtype
            else:
                rql = "SET Y %s X WHERE X eid %%(x)s, Y eid %%(y)s" % rtype
            for teid in eids:
                req.execute(rql, {"x": eid, "y": int(teid)})


def _validation_error(req, ex):
    req.cnx.rollback()
    ex.translate(req._)  # translate messages using ui language
    # XXX necessary to remove existant validation error?
    # imo (syt), it's not necessary
    req.session.data.pop(req.form.get("__errorurl"), None)
    foreid = ex.entity
    eidmap = req.data.get("eidmap", {})
    for var, eid in eidmap.items():
        if foreid == eid:
            foreid = var
            break
    return (foreid, ex.errors)


def _validate_form(req, vreg):
    """
    helper function for FormValidatorController.publish()

    return (status, args, entity) that will be passed as argument
    result to `cubicweb.edition.js:handleFormValidationResult(..., result, ...)`

    status: bool True if valid else False
    args: target location if status is True else error message
    entity: entity being edited
    """
    # XXX should use the `RemoteCallFailed` mechanism
    try:
        ctrl = vreg["controllers"].select("edit", req=req)
    except NoSelectableObject:
        return (False, {None: req._("not authorized")}, None)
    try:
        ctrl.publish(None)
    except ValidationError as ex:
        return (False, _validation_error(req, ex), ctrl._edited_entity)
    except Redirect as ex:
        try:
            txuuid = req.cnx.commit()  # ValidationError may be raised on commit
        except ValidationError as ex:
            return (False, _validation_error(req, ex), ctrl._edited_entity)
        except Exception as ex:
            req.cnx.rollback()
            req.exception("unexpected error while validating form")
            return (False, str(ex), ctrl._edited_entity)
        else:
            if txuuid is not None:
                req.data["last_undoable_transaction"] = txuuid
            # complete entity: it can be used in js callbacks where we might
            # want every possible information
            if ctrl._edited_entity:
                ctrl._edited_entity.complete()
            return (True, ex.location, ctrl._edited_entity)
    except Exception as ex:
        req.cnx.rollback()
        req.exception("unexpected error while validating form")
        return (False, str(ex), ctrl._edited_entity)
    return (False, "???", None)


class FormValidatorController(Controller):
    __regid__ = "validateform"

    def response(self, domid, status, args, entity):
        callback = str(self._cw.form.get("__onsuccess", "null"))
        errback = str(self._cw.form.get("__onfailure", "null"))
        cbargs = str(self._cw.form.get("__cbargs", "null"))
        self._cw.set_content_type("text/html")
        jsargs = json_dumps((status, args, entity))
        return """<script type="text/javascript">
 window.parent.handleFormValidationResponse('%s', %s, %s, %s, %s);
</script>""" % (
            domid,
            callback,
            errback,
            jsargs,
            cbargs,
        )

    def publish(self, rset=None):
        self._cw.ajax_request = True
        # XXX unclear why we have a separated controller here vs
        # js_validate_form on the json controller
        status, args, entity = _validate_form(self._cw, self._cw.vreg)
        domid = self._cw.form.get("__domid", "entityForm")
        return self.response(domid, status, args, entity).encode(self._cw.encoding)


class MailBugReportController(Controller):
    __regid__ = "reportbug"
    __select__ = match_form_params("description")

    def publish(self, rset=None):
        req = self._cw
        desc = req.form["description"]
        # The description is generated and signed by cubicweb itself, check
        # description's signature so we don't want to send spam here
        sign = req.form.get("__signature", "")
        if not (sign and req.vreg.config.check_text_sign(desc, sign)):
            raise Forbidden("Invalid content")
        self.sendmail(
            req.vreg.config["submit-mail"],
            req._("%s error report") % req.vreg.config.appid,
            desc,
        )
        raise Redirect(req.build_url(__message=req._("bug report sent")))


class UndoController(Controller):
    __regid__ = "undo"
    __select__ = authenticated_user() & match_form_params("txuuid")

    def publish(self, rset=None):
        txuuid = self._cw.form["txuuid"]
        try:
            self._cw.cnx.undo_transaction(txuuid)
        except UndoTransactionException as exc:
            errors = exc.errors
            # This will cause a rollback in main_publish
            raise ValidationError(None, {None: "\n".join(errors)})
        else:
            self.redirect()  # Will raise Redirect

    def redirect(self, msg=None):
        req = self._cw
        msg = msg or req._("transaction undone")
        self._redirect({"_cwmsgid": req.set_redirect_message(msg)})



from urllib.parse import urljoin

from cubicweb.devtools import BASE_URL
from cubicweb_web.devtools.testlib import PyramidWebCWTC
from cubicweb_web.views.basecontrollers import ViewController


class ControllerWithCSRFCheckDisabled(ViewController):
    require_csrf = False


class CustomControllerWithoutCSRFTest(PyramidWebCWTC):
    def test_controller_without_csrf(self):
        self.webapp.login()
        url = urljoin(BASE_URL, "cwuser", self.webapp.admin_login)
        kwargs = {
            "do_not_grab_the_crsf_token": True,
            "do_not_inject_origin": True,
        }

        # failed with 400 because we didn't send the csrf token
        self.webapp.post(
            url,
            params={},
            status=400,
            **kwargs,
        )

        # activate the custom controller that has require_csrf = False
        self.vreg.register_and_replace(ControllerWithCSRFCheckDisabled, ViewController)

        # request works without the csrf token
        self.webapp.post(
            url,
            params="test",
            **kwargs,
        )
        self.webapp.post(
            url,
            params={},
            **kwargs,
        )
        self.webapp.post(
            url,
            params=[],
            **kwargs,
        )

        # request works without the csrf token
        self.webapp.post_json(
            url,
            params={},
            **kwargs,
        )
        self.webapp.post_json(
            url,
            params=[],
            **kwargs,
        )

        # request works without the csrf token
        self.webapp.put_json(
            url,
            params={},
            **kwargs,
        )


class CSRFTest(PyramidWebCWTC):
    def test_pyramid_route_csrf_token_is_present(self):
        res = self.webapp.get(urljoin(BASE_URL, "login"))
        self.assertIn("csrf_token", res.form.fields)

    def test_pyramid_route_csrf_bad_token(self):
        self.webapp.post(
            urljoin(BASE_URL, "login"),
            {
                "__login": self.admlogin,
                "__password": self.admpassword,
                "csrf_token": "bad_token",
            },
            status=400,
        )

    def test_pyramid_route_csrf_no_token(self):
        self.webapp.post(
            urljoin(BASE_URL, "login"),
            {
                "__login": self.admlogin,
                "__password": self.admpassword,
                "csrf_token": None,
            },
            status=400,
        )

    def test_pyramid_route_csrf_bad_origin(self):
        self.webapp.post(
            urljoin(BASE_URL, "login"),
            {"__login": self.admlogin, "__password": self.admpassword},
            headers={"Origin": "bad_origin.net"},
            status=400,
        )

    def test_pyramid_route_csrf_no_origin(self):
        self.webapp.post(
            urljoin(BASE_URL, "login"),
            {"__login": self.admlogin, "__password": self.admpassword},
            do_not_inject_origin=True,
            status=400,
        )

    def test_cubicweb_route_csrf_token_is_present(self):
        self.webapp.post(
            urljoin(BASE_URL, "validateform"),
            {
                "__form_id": "edition",
                "__type:6": "CWUser",
                "eid": 6,
                "firstname-subject:6": "loutre",
            },
        )

    def test_cubicweb_route_no_csrf_token(self):
        self.webapp.post(
            urljoin(BASE_URL, "validateform"),
            {
                "__form_id": "edition",
                "__type:6": "CWUser",
                "eid": 6,
                "firstname-subject:6": "loutre",
                "csrf_token": None,
            },
            status=400,
        )

    def test_cubicweb_route_bad_origin(self):
        self.webapp.post(
            urljoin(BASE_URL, "validateform"),
            {
                "__form_id": "edition",
                "__type:6": "CWUser",
                "eid": 6,
                "firstname-subject:6": "loutre",
            },
            headers={"Origin": "bad_origin.net"},
            status=400,
        )

    def test_cubicweb_route_csrf_no_origin(self):
        self.webapp.post(
            urljoin(BASE_URL, "validateform"),
            {
                "__form_id": "edition",
                "__type:6": "CWUser",
                "eid": 6,
                "firstname-subject:6": "loutre",
            },
            do_not_inject_origin=True,
            status=400,
        )

    def test_pyramid_route_csrf_token_post_json(self):
        self.webapp.login()
        self.webapp.post_json(
            urljoin(BASE_URL, "cwuser", self.webapp.admin_login),
            params={},
            status=200,
        )

    def test_pyramid_route_csrf_token_put_json(self):
        self.webapp.login()
        self.webapp.put_json(
            urljoin(BASE_URL, "cwuser", self.webapp.admin_login),
            params={},
            status=200,
        )



import json
import re
import sys
from contextlib import contextmanager
from math import log
from urllib.parse import urlparse, parse_qs, unquote, urljoin

from cubicweb.devtools import (
    DEFAULT_EMPTY_DB_ID,
    VIEW_VALIDATORS,
    BASE_URL,
    SYSTEM_RELATIONS,
)
from cubicweb.devtools.apptest_config import (
    ApptestConfiguration,
    PostgresApptestConfiguration,
)
from cubicweb.devtools.fake import FakeCWRegistryStore, FakeConfig
from cubicweb.devtools.fill import insert_entity_queries, make_relations_queries
from cubicweb.devtools.htmlparser import (
    XMLValidator,
    VALMAP,
    XMLSyntaxValidator,
    DTDValidator,
    HTMLValidator,
)
from cubicweb.devtools.testlib import (
    CubicWebTC,
    RepoAccess,
    Session,
    not_selected,
    JsonValidator,
    line_context_filter,
    real_error_handling,
    unprotected_entities,
    CubicWebDebugger,
)
from cubicweb.pyramid.test import PyramidCWTest, ACCEPTED_ORIGINS, CustomTestApp
from cubicweb.sobjects.notification import NotificationView
from cubicweb.uilib import eid_param
from logilab.common.decorators import cachedproperty
from logilab.common.deprecation import class_deprecated
from logilab.common.registry import NoSelectableObject
from logilab.common.testlib import Tags, nocoverage
from pyramid.csrf import LegacySessionCSRFStoragePolicy
from pyramid.interfaces import ICSRFStoragePolicy
from pyramid.testing import DummyRequest
from webtest import AppError
from yams import ValidationError

from cubicweb_web._exceptions import Redirect
from cubicweb_web.application import CubicWebPublisher
from cubicweb_web.request import CubicWebRequestBase
from cubicweb_web.webconfig import WebAllInOneConfiguration


class WebApptestConfiguration(ApptestConfiguration, WebAllInOneConfiguration):
    pass


class WebPostgresApptestConfiguration(
    PostgresApptestConfiguration, WebAllInOneConfiguration
):
    pass


class FakeRequest(CubicWebRequestBase):
    """test implementation of an cubicweb request object"""

    def __init__(self, *args, **kwargs):
        if not (args or "vreg" in kwargs):
            kwargs["vreg"] = FakeCWRegistryStore(FakeConfig(), initlog=False)
        self._http_method = kwargs.pop("method", "GET")
        self._url = kwargs.pop("url", None)
        if self._url is None:
            self._url = "view?rql=Blop&vid=blop"
        super().__init__(*args, **kwargs)
        self._session_data = {}
        self._request = DummyRequest()
        self._csrf_storage = LegacySessionCSRFStoragePolicy()
        self._request.registry.registerUtility(
            self._csrf_storage, provided=ICSRFStoragePolicy
        )

    def set_cookie(
        self, name, value, maxage=300, expires=None, secure=False, httponly=False
    ):
        super().set_cookie(name, value, maxage, expires, secure, httponly)
        cookie = self.get_response_header("Set-Cookie")
        self._headers_in.setHeader("Cookie", cookie)

    # Implement request abstract API
    def http_method(self):
        return self._http_method

    def relative_path(self, includeparams=True):
        """return the normalized path of the request (ie at least relative
        to the instance's root, but some other normalization may be needed
        so that the returned path may be used to compare to generated urls
        """
        if self._url.startswith(BASE_URL):
            url = self._url[len(BASE_URL) :]
        else:
            url = self._url
        if includeparams:
            return url
        return url.split("?", 1)[0]

    def set_request_header(self, header, value, raw=False):
        """set an incoming HTTP header (for test purpose only)"""
        if isinstance(value, str):
            value = [value]
        if raw:
            # adding encoded header is important, else page content
            # will be reconverted back to unicode and apart unefficiency, this
            # may cause decoding problem (e.g. when downloading a file)
            self._headers_in.setRawHeaders(header, value)
        else:
            self._headers_in.setHeader(header, value)  #

    def get_response_header(self, header, default=None, raw=False):
        """return output header (for test purpose only)"""
        if raw:
            return self.headers_out.getRawHeaders(header, [default])[0]
        return self.headers_out.getHeader(header, default)

    def build_url_params(self, **kwargs):
        # overriden to get predictable resultts
        args = []
        for param, values in sorted(kwargs.items()):
            if not isinstance(values, (list, tuple)):
                values = (values,)
            for value in values:
                assert value is not None
                args.append("{}={}".format(param, self.url_quote(value)))
        return "&".join(args)


class WebRepoAccess(RepoAccess):
    def __init__(self, repo, login, requestcls):
        super().__init__(repo, login)
        self.requestcls = requestcls

    @contextmanager
    def web_request(self, url=None, headers={}, method="GET", **kwargs):
        """Context manager returning a web request pre-linked to a client cnx

        To commit and rollback use::

            req.cnx.commit()
            req.cnx.rolback()
        """
        session = kwargs.pop("session", Session(self._repo, self._user))
        req = self.requestcls(
            self._repo.vreg, url=url, headers=headers, method=method, form=kwargs
        )
        with self.cnx() as cnx:
            # web request expect a session attribute on cnx referencing the web session
            cnx.session = session
            req.set_cnx(cnx)
            yield req


class WebCWTC(CubicWebTC):
    """abstract class for test using an apptest environment

    attributes:

    * `vreg`, the vregistry
    * `schema`, self.vreg.schema
    * `config`, cubicweb configuration
    * `cnx`, repoapi connection to the repository using an admin user
    * `session`, server side session associated to `cnx`
    * `app`, the cubicweb publisher (for web testing)
    * `repo`, the repository object
    * `admlogin`, login of the admin user
    * `admpassword`, password of the admin user
    * `shell`, create and use shell environment
    * `anonymous_allowed`: flag telling if anonymous browsing should be allowed
    """

    appid = "data"
    configcls = WebApptestConfiguration
    requestcls = FakeRequest
    tags = Tags("cubicweb", "cw_repo")
    test_db_id = DEFAULT_EMPTY_DB_ID

    # anonymous is logged by default in cubicweb test cases
    anonymous_allowed = True

    def __init__(self, *args, **kwargs):
        self.repo = None
        self._open_access = set()
        super(CubicWebTC, self).__init__(*args, **kwargs)

    # repository connection handling ###########################################

    def new_access(self, login):
        """provide a new RepoAccess object for a given user

        The access is automatically closed at the end of the test."""
        access = WebRepoAccess(self.repo, login, self.requestcls)
        self._open_access.add(access)
        return access

    def pviews(self, req, rset):
        return sorted(
            (a.__regid__, a.__class__)
            for a in self.vreg["views"].possible_views(req, rset=rset)
        )

    def pactions(
        self,
        req,
        rset,
        skipcategories=("addrelated", "siteactions", "useractions", "footer", "manage"),
    ):
        return [
            (a.__regid__, a.__class__)
            for a in self.vreg["actions"].poss_visible_objects(req, rset=rset)
            if a.category not in skipcategories
        ]

    def pactions_by_cats(self, req, rset, categories=("addrelated",)):
        return [
            (a.__regid__, a.__class__)
            for a in self.vreg["actions"].poss_visible_objects(req, rset=rset)
            if a.category in categories
        ]

    def pactionsdict(
        self,
        req,
        rset,
        skipcategories=("addrelated", "siteactions", "useractions", "footer", "manage"),
    ):
        res = {}
        for a in self.vreg["actions"].poss_visible_objects(req, rset=rset):
            if a.category not in skipcategories:
                res.setdefault(a.category, []).append(a.__class__)
        return res

    def action_submenu(self, req, rset, id):
        return self._test_action(self.vreg["actions"].select(id, req, rset=rset))

    def _test_action(self, action):
        class fake_menu(list):
            @property
            def items(self):
                return self

        class fake_box(object):
            def action_link(self, action, **kwargs):
                return (action.title, action.url())

        submenu = fake_menu()
        action.fill_menu(fake_box(), submenu)
        return submenu

    def list_views_for(self, rset):
        """returns the list of views that can be applied on `rset`"""
        req = rset.req
        only_once_vids = ("primary", "secondary", "text")
        req.data["ex"] = ValueError("whatever")
        viewsvreg = self.vreg["views"]
        for vid, views in viewsvreg.items():
            if vid[0] == "_":
                continue
            if rset.rowcount > 1 and vid in only_once_vids:
                continue
            views = [
                view
                for view in views
                if view.category != "startupview"
                and not issubclass(view, NotificationView)
                and not isinstance(view, class_deprecated)
            ]
            if views:
                try:
                    view = viewsvreg._select_best(views, req, rset=rset)
                    if view is None:
                        raise NoSelectableObject((req,), {"rset": rset}, views)
                    if view.linkable():
                        yield view
                    else:
                        not_selected(self.vreg, view)
                    # else the view is expected to be used as subview and should
                    # not be tested directly
                except NoSelectableObject:
                    continue

    def list_actions_for(self, rset):
        """returns the list of actions that can be applied on `rset`"""
        req = rset.req
        for action in self.vreg["actions"].possible_objects(req, rset=rset):
            yield action

    def list_boxes_for(self, rset):
        """returns the list of boxes that can be applied on `rset`"""
        req = rset.req
        for box in self.vreg["ctxcomponents"].possible_objects(
            req, rset=rset, view=None
        ):
            yield box

    def list_startup_views(self):
        """returns the list of startup views"""
        with self.admin_access.web_request() as req:
            for view in self.vreg["views"].possible_views(req, None):
                if view.category == "startupview":
                    yield view.__regid__
                else:
                    not_selected(self.vreg, view)

    # web ui testing utilities #################################################

    @cachedproperty
    def app(self):
        """return a cubicweb publisher"""
        publisher = CubicWebPublisher(self.repo, self.config)

        def raise_error_handler(*args, **kwargs):
            raise

        publisher.error_handler = raise_error_handler
        return publisher

    @contextmanager
    def remote_calling(self, fname, *args, **kwargs):
        """remote json call simulation"""
        args = [json.dumps(arg) for arg in args]
        with self.admin_access.web_request(
            fname=fname, pageid="123", arg=args, **kwargs
        ) as req:
            ctrl = self.vreg["controllers"].select("ajax", req)
            yield ctrl.publish(), req

    def app_handle_request(self, req):
        return self.app.core_handle(req)

    def ctrl_publish(self, req, ctrl="edit", rset=None):
        """call the publish method of the edit controller"""
        ctrl = self.vreg["controllers"].select(ctrl, req, appli=self.app)
        try:
            result = ctrl.publish(rset)
            req.cnx.commit()
        except Redirect:
            req.cnx.commit()
            raise
        return result

    @staticmethod
    def fake_form(formid, field_dict=None, entity_field_dicts=()):
        """Build _cw.form dictionnary to fake posting of some standard cubicweb form

        * `formid`, the form id, usually form's __regid__

        * `field_dict`, dictionary of name:value for fields that are not tied to an entity

        * `entity_field_dicts`, list of (entity, dictionary) where dictionary contains name:value
          for fields that are not tied to the given entity
        """
        assert (
            field_dict or entity_field_dicts
        ), "field_dict and entity_field_dicts arguments must not be both unspecified"
        if field_dict is None:
            field_dict = {}
        form = {"__form_id": formid}
        fields = []
        for field, value in field_dict.items():
            fields.append(field)
            form[field] = value

        def _add_entity_field(entity, field, value):
            entity_fields.append(field)
            form[eid_param(field, entity.eid)] = value

        for entity, field_dict in entity_field_dicts:
            if "__maineid" not in form:
                form["__maineid"] = entity.eid
            entity_fields = []
            form.setdefault("eid", []).append(entity.eid)
            _add_entity_field(entity, "__type", entity.cw_etype)
            for field, value in field_dict.items():
                _add_entity_field(entity, field, value)
            if entity_fields:
                form[eid_param("_cw_entity_fields", entity.eid)] = ",".join(
                    entity_fields
                )
        if fields:
            form["_cw_fields"] = ",".join(sorted(fields))
        return form

    @contextmanager
    def admin_request_from_url(self, url):
        """parses `url` and builds the corresponding CW-web request

        req.form will be setup using the url's query string
        """
        with self.admin_access.web_request(url=url) as req:
            if isinstance(url, str):
                url = url.encode(
                    req.encoding
                )  # req.setup_params() expects encoded strings
            querystring = urlparse(url)[-2]
            params = parse_qs(querystring)
            req.setup_params(params)
            yield req

    @staticmethod
    def _parse_location(req, location):
        try:
            path, params = location.split("?", 1)
        except ValueError:
            path = location
            params = {}
        else:

            def cleanup(p):
                return (p[0], unquote(p[1]))

            params = dict(cleanup(p.split("=", 1)) for p in params.split("&") if p)
        if path.startswith(req.base_url()):  # may be relative
            path = path[len(req.base_url()) :]
        return path, params

    def expect_redirect(self, callback, req):
        """call the given callback with req as argument, expecting to get a
        Redirect exception
        """
        try:
            callback(req)
        except Redirect as ex:
            return self._parse_location(req, ex.location)
        else:
            self.fail("expected a Redirect exception")

    def expect_redirect_handle_request(self, req, path="edit"):
        """call the publish method of the application publisher, expecting to
        get a Redirect exception
        """
        if req.relative_path(False) != path:
            req._url = path
        self.app_handle_request(req)
        self.assertTrue(300 <= req.status_out < 400, req.status_out)
        location = req.get_response_header("location")
        return self._parse_location(req, location)

    # content validation #######################################################

    # validators are used to validate (XML, DTD, whatever) view's content
    # validators availables are :
    #  DTDValidator : validates XML + declared DTD
    #  SaxOnlyValidator : guarantees XML is well formed
    #  None : do not try to validate anything
    # validators used must be imported from from.devtools.htmlparser
    content_type_validators = {
        # maps MIME type : validator name
        #
        # do not set html validators here, we need HTMLValidator for html
        # snippets
        # 'text/html': DTDValidator,
        # 'application/xhtml+xml': DTDValidator,
        "application/xml": XMLValidator,
        "text/xml": XMLValidator,
        "application/json": JsonValidator,
        "text/plain": None,
        "text/comma-separated-values": None,
        "text/x-vcard": None,
        "text/calendar": None,
        "image/png": None,
    }
    # maps vid : validator name (override content_type_validators)
    vid_validators = dict(
        (vid, VALMAP[valkey]) for vid, valkey in VIEW_VALIDATORS.items()
    )

    def view(self, vid, rset=None, req=None, template="main-template", **kwargs):
        """This method tests the view `vid` on `rset` using `template`

        If no error occurred while rendering the view, the HTML is analyzed
        and parsed.

        :returns: an instance of `cubicweb.devtools.htmlparser.PageInfo`
                  encapsulation the generated HTML
        """
        if req is None:
            assert rset is not None, "you must supply at least one of rset or req"
            req = rset.req
        req.form["vid"] = vid
        viewsreg = self.vreg["views"]
        view = viewsreg.select(vid, req, rset=rset, **kwargs)
        if template is None:  # raw view testing, no template
            viewfunc = view.render
        else:
            kwargs["view"] = view

            def viewfunc(**k):
                return viewsreg.main_template(req, template, rset=rset, **kwargs)

        return self._test_view(viewfunc, view, template, kwargs)

    def _test_view(self, viewfunc, view, template="main-template", kwargs={}):
        """this method does the actual call to the view

        If no error occurred while rendering the view, the HTML is analyzed
        and parsed.

        :returns: an instance of `cubicweb.devtools.htmlparser.PageInfo`
                  encapsulation the generated HTML
        """
        try:
            output = viewfunc(**kwargs)
        except Exception:
            # hijack exception: generative tests stop when the exception
            # is not an AssertionError
            klass, exc, tcbk = sys.exc_info()
            try:
                msg = "[%s in %s] %s" % (klass, view.__regid__, exc)
            except Exception:
                msg = "[%s in %s] undisplayable exception" % (klass, view.__regid__)
            raise AssertionError(msg).with_traceback(sys.exc_info()[-1])
        return self._check_html(output, view, template)

    def get_validator(self, view=None, content_type=None, output=None):
        if view is not None:
            try:
                return self.vid_validators[view.__regid__]()
            except KeyError:
                if content_type is None:
                    content_type = view.content_type
        if content_type is None:
            content_type = "text/html"
        if content_type in ("text/html", "application/xhtml+xml") and output:
            if output.startswith(b"<!DOCTYPE html>"):
                # only check XML well-formness since HTMLValidator isn't html5
                # compatible and won't like various other extensions
                default_validator = XMLSyntaxValidator
            elif output.startswith(b"<?xml"):
                default_validator = DTDValidator
            else:
                default_validator = HTMLValidator
        else:
            default_validator = None
        validatorclass = self.content_type_validators.get(
            content_type, default_validator
        )
        if validatorclass is None:
            return
        return validatorclass()

    @nocoverage
    def _check_html(self, output, view, template="main-template"):
        """raises an exception if the HTML is invalid"""
        output = output.strip()
        if isinstance(output, str):
            # XXX
            output = output.encode("utf-8")
        validator = self.get_validator(view, output=output)
        if validator is None:
            return output  # return raw output if no validator is defined
        if isinstance(validator, DTDValidator):
            # XXX remove <canvas> used in progress widget, unknown in html dtd
            output = re.sub("<canvas.*?></canvas>", "", output)
        return self.assertWellFormed(validator, output.strip(), context=view)

    def assertWellFormed(self, validator, content, context=None):
        try:
            return validator.parse_string(content)
        except Exception:
            # hijack exception: generative tests stop when the exception
            # is not an AssertionError
            klass, exc, tcbk = sys.exc_info()
            if context is None:
                msg = "[%s]" % (klass,)
            else:
                msg = "[%s in %s view] " % (klass, context)
            msg = msg.encode(sys.getdefaultencoding(), "replace")

            try:
                str_exc = str(exc)
            except Exception:
                str_exc = "undisplayable exception"
            msg += str_exc.encode(sys.getdefaultencoding(), "replace")
            if content is not None:
                position = getattr(exc, "position", (0,))[0]
                if position:
                    # define filter
                    if isinstance(content, str):
                        content = content.encode(sys.getdefaultencoding(), "replace")
                    content = validator.preprocess_data(content)
                    content = content.splitlines()
                    width = int(log(len(content), 10)) + 1
                    line_template = b" %" + (b"%i" % width) + b"i: %s"
                    error_line_template = b">%" + (b"%i" % width) + b"i: %s"
                    # XXX no need to iterate the whole file except to get
                    # the line number
                    content = b"\n".join(
                        (
                            line_template
                            if (idx + 1) != position
                            else error_line_template
                        )
                        % (idx + 1, line)
                        for idx, line in enumerate(content)
                        if line_context_filter(idx + 1, position)
                    )
                    msg += b"\nfor content:\n%s" % content
            exc = AssertionError(msg.decode())
            exc.__traceback__ = tcbk
            raise exc

    def assertDocTestFile(self, testfile):
        # doctest returns tuple (failure_count, test_count)
        with self.admin_access.shell() as mih:
            result = mih.process_script(testfile)
        if result[0] and result[1]:
            raise self.failureException("doctest file '%s' failed" % testfile)


class PyramidWebTestApp(CustomTestApp):
    def __init__(self, *args, admin_login, admin_password, **kwargs):
        super().__init__(*args, **kwargs)
        self.admin_login = admin_login
        self.admin_password = admin_password
        self._ident_cookie = None
        self._csrf_token = None

    def reset(self):
        super().reset()
        self._ident_cookie = None
        self._csrf_token = None

    def post(
        self,
        route,
        params=None,
        do_not_grab_the_crsf_token=False,
        do_not_inject_origin=False,
        **kwargs,
    ):
        if params is None:
            params = {}

        if (
            isinstance(params, dict)
            and not do_not_grab_the_crsf_token
            and "csrf_token" not in params
        ):
            csrf_token = self.get_csrf_token()

            # "application/json" doesn't submit token in form params but as header value
            if kwargs.get("headers", {}).get("Content-Type") != "application/json":
                if "csrf_token" not in params:
                    params["csrf_token"] = csrf_token
            else:
                if "headers" in kwargs:
                    kwargs["headers"]["X-CSRF-Token"] = csrf_token
                else:
                    kwargs["headers"] = {"X-CSRF-Token": csrf_token}

        if not do_not_inject_origin:
            if "headers" in kwargs and "Origin" not in kwargs["headers"]:
                kwargs["headers"]["Origin"] = "https://" + ACCEPTED_ORIGINS[0]
            elif "headers" not in kwargs:
                kwargs["headers"] = {"Origin": "https://" + ACCEPTED_ORIGINS[0]}

        return super().post(route, params, **kwargs)

    def _inject_params_to_kwargs(
        self, do_not_grab_the_crsf_token, do_not_inject_origin, **kwargs
    ):
        if not do_not_grab_the_crsf_token:
            csrf_token = self.get_csrf_token()
            if "headers" in kwargs:
                kwargs["headers"]["X-CSRF-Token"] = csrf_token
            else:
                kwargs["headers"] = {"X-CSRF-Token": csrf_token}
        if not do_not_inject_origin:
            if "headers" in kwargs and "Origin" not in kwargs["headers"]:
                kwargs["headers"]["Origin"] = "https://" + ACCEPTED_ORIGINS[0]
            elif "headers" not in kwargs:
                kwargs["headers"] = {"Origin": "https://" + ACCEPTED_ORIGINS[0]}
        return kwargs

    def post_json(
        self,
        route,
        params=None,
        do_not_grab_the_crsf_token=False,
        do_not_inject_origin=False,
        **kwargs,
    ):
        kwargs = self._inject_params_to_kwargs(
            do_not_grab_the_crsf_token, do_not_inject_origin, **kwargs
        )
        return super().post_json(route, params, **kwargs)

    def put_json(
        self,
        route,
        params=None,
        do_not_grab_the_crsf_token=False,
        do_not_inject_origin=False,
        **kwargs,
    ):
        kwargs = self._inject_params_to_kwargs(
            do_not_grab_the_crsf_token, do_not_inject_origin, **kwargs
        )
        return super().put_json(route, params, **kwargs)

    def get_csrf_token(self):
        if "csrf_token" in self.cookies:
            return self.cookies["csrf_token"]

        try:
            self.get(BASE_URL)
        except AppError as e:
            # url doesn't exist or forbidden
            if "404" in e.args[0] or "403" in e.args[0]:
                try:
                    response = self.get(urljoin(BASE_URL, "login"))
                except Exception as e:
                    sys.stderr.write(
                        f"ERROR: failed to do a GET on '/' and on '/login' to get the csrf token because: {e}\n"
                    )
                    raise
                else:
                    if "csrf_token" in self.cookies:
                        return self.cookies["csrf_token"]
                    return response.form.fields["csrf_token"][0].value
            else:
                sys.stderr.write(
                    f"ERROR: failed to do a GET on '/' to get the csrf token because: {e}\n"
                )
                raise
        else:
            return self.cookies["csrf_token"]

    def login(self, user=None, password=None, status=303, additional_arguments=None):
        """Log the current http session for the provided credential

        If no user is provided, admin connection are used.
        """
        if user is None:
            user = self.admin_login
            password = self.admin_password
        if password is None:
            password = user

        arguments = {"__login": user, "__password": password}
        if additional_arguments:
            arguments.update(additional_arguments)

        response = self.post(urljoin(BASE_URL, "login"), arguments, status=status)

        assert response.status_int == status

        return response

    def logout(self):
        response = self.get(urljoin(BASE_URL, "logout"))
        self.reset()
        return response


class PyramidWebCWTC(PyramidCWTest, WebCWTC):
    settings = {
        "cubicweb.includes": ["cubicweb.pyramid.auth", "cubicweb.pyramid.session"]
    }

    def includeme(self, config):
        PyramidCWTest.includeme(self, config)
        config.registry.settings["pyramid.csrf_trusted_origins"] = ACCEPTED_ORIGINS

    def url_publish(self, url, data=None):
        """takes `url`, uses application's app_resolver to find the appropriate
        controller and result set, then publishes the result.

        To simulate post of www-form-encoded data, give a `data` dictionary
        containing desired key/value associations.

        This should pretty much correspond to what occurs in a real CW server
        except the apache-rewriter component is not called.
        """
        with self.admin_request_from_url(url) as req:
            if data is not None:
                req.form.update(data)
            ctrlid, rset = self.app.url_resolver.process(req, req.relative_path(False))
            return self.ctrl_publish(req, ctrlid, rset)

    def http_publish(self, url, data=None):
        """like `url_publish`, except this returns a http response, even in case
        of errors. You may give form parameters using the `data` argument.
        """
        with self.admin_request_from_url(url) as req:
            if data is not None:
                req.form.update(data)
            with real_error_handling(self.app):
                result = self.app_handle_request(req)
            return result, req

    def build_webapp(self):
        self.webapp = PyramidWebTestApp(
            self.pyramid_config.make_wsgi_app(),
            extra_environ={"wsgi.url_scheme": "https"},
            admin_login=self.admlogin,
            admin_password=self.admpassword,
        )

    def tearDown(self):
        del self.webapp
        super().tearDown()


def how_many_dict(schema, cnx, how_many, skip):
    """given a schema, compute how many entities by type we need to be able to
    satisfy relations cardinality.

    The `how_many` argument tells how many entities of which type we want at
    least.

    Return a dictionary with entity types as key, and the number of entities for
    this type as value.
    """
    relmap = {}
    for rschema in schema.relations():
        if rschema.final:
            continue
        for subj, obj in rschema.relation_definitions:
            card = rschema.relation_definition(subj, obj).cardinality
            # if the relation is mandatory, we'll need at least as many subj and
            # obj to satisfy it
            if card[0] in "1+" and card[1] in "1?":
                # subj has to be linked to at least one obj,
                # but obj can be linked to only one subj
                # -> we need at least as many subj as obj to satisfy
                #    cardinalities for this relation
                relmap.setdefault((rschema, subj), []).append(str(obj))
            if card[1] in "1+" and card[0] in "1?":
                # reverse subj and obj in the above explanation
                relmap.setdefault((rschema, obj), []).append(str(subj))
    unprotected = unprotected_entities(schema)
    for etype in skip:  # XXX (syt) duh? explain or kill
        unprotected.add(etype)
    howmanydict = {}
    # step 1, compute a base number of each entity types: number of already
    # existing entities of this type + `how_many`
    for etype in unprotected_entities(schema, strict=True):
        howmanydict[str(etype)] = cnx.execute("Any COUNT(X) WHERE X is %s" % etype)[0][
            0
        ]
        if etype in unprotected:
            howmanydict[str(etype)] += how_many
    # step 2, augment nb entity per types to satisfy cardinality constraints,
    # by recomputing for each relation that constrained an entity type:
    #
    # new num for etype = max(current num, sum(num for possible target etypes))
    #
    # XXX we should first check there is no cycle then propagate changes
    for (rschema, etype), targets in relmap.items():
        relfactor = sum(howmanydict[e] for e in targets)
        howmanydict[str(etype)] = max(relfactor, howmanydict[etype])
    return howmanydict


class AutoPopulateTest(WebCWTC):
    """base class for test with auto-populating of the database"""

    __abstract__ = True

    test_db_id = "autopopulate"

    tags = WebCWTC.tags | Tags("autopopulated")

    pdbclass = CubicWebDebugger
    # this is a hook to be able to define a list of rql queries
    # that are application dependent and cannot be guessed automatically
    application_rql = []

    no_auto_populate = ()
    ignored_relations = set()

    def to_test_etypes(self):
        return unprotected_entities(self.schema, strict=True)

    def custom_populate(self, how_many, cnx):
        pass

    def post_populate(self, cnx):
        pass

    @nocoverage
    def auto_populate(self, how_many):
        """this method populates the database with `how_many` entities
        of each possible type. It also inserts random relations between them
        """
        with self.admin_access.cnx() as cnx:
            with cnx.security_enabled(read=False, write=False):
                self._auto_populate(cnx, how_many)
                cnx.commit()

    def _auto_populate(self, cnx, how_many):
        self.custom_populate(how_many, cnx)
        vreg = self.vreg
        howmanydict = how_many_dict(self.schema, cnx, how_many, self.no_auto_populate)
        for etype in unprotected_entities(self.schema):
            if etype in self.no_auto_populate:
                continue
            nb = howmanydict.get(etype, how_many)
            for rql, args in insert_entity_queries(etype, self.schema, vreg, nb):
                cnx.execute(rql, args)
        edict = {}
        for etype in unprotected_entities(self.schema, strict=True):
            rset = cnx.execute("%s X" % etype)
            edict[str(etype)] = set(row[0] for row in rset.rows)
        existingrels = {}
        ignored_relations = SYSTEM_RELATIONS | self.ignored_relations
        for rschema in self.schema.relations():
            if rschema.final or rschema in ignored_relations or rschema.rule:
                continue
            rset = cnx.execute("DISTINCT Any X,Y WHERE X %s Y" % rschema)
            existingrels.setdefault(rschema.type, set()).update((x, y) for x, y in rset)
        q = make_relations_queries(
            self.schema, edict, cnx, ignored_relations, existingrels=existingrels
        )
        for rql, args in q:
            try:
                cnx.execute(rql, args)
            except ValidationError as ex:
                # failed to satisfy some constraint
                print("error in automatic db population", ex)
                cnx.commit_state = None  # reset uncommitable flag
        self.post_populate(cnx)

    def iter_individual_rsets(self, etypes=None, limit=None):
        etypes = etypes or self.to_test_etypes()
        with self.admin_access.web_request() as req:
            for etype in etypes:
                if limit:
                    rql = "Any X LIMIT %s WHERE X is %s" % (limit, etype)
                else:
                    rql = "Any X WHERE X is %s" % etype
                rset = req.execute(rql)
                for row in range(len(rset)):
                    if limit and row > limit:
                        break
                    # XXX iirk
                    rset2 = rset.limit(limit=1, offset=row)
                    yield rset2

    def iter_automatic_rsets(self, limit=10):
        """generates basic resultsets for each entity type"""
        etypes = self.to_test_etypes()
        if not etypes:
            return
        with self.admin_access.web_request() as req:
            for etype in etypes:
                yield req.execute("Any X LIMIT %s WHERE X is %s" % (limit, etype))
            etype1 = etypes.pop()
            try:
                etype2 = etypes.pop()
            except KeyError:
                etype2 = etype1
            # test a mixed query (DISTINCT/GROUP to avoid getting duplicate
            # X which make muledit view failing for instance (html validation fails
            # because of some duplicate "id" attributes)
            yield req.execute(
                "DISTINCT Any X, MAX(Y) GROUPBY X WHERE X is %s, Y is %s"
                % (etype1, etype2)
            )
            # test some application-specific queries if defined
            for rql in self.application_rql:
                yield req.execute(rql)

    def _test_everything_for(self, rset):
        """this method tries to find everything that can be tested
        for `rset` and yields a callable test (as needed in generative tests)
        """
        propdefs = self.vreg["propertydefs"]
        # make all components visible
        for k, v in propdefs.items():
            if k.endswith("visible") and not v["default"]:
                propdefs[k]["default"] = True
        for view in self.list_views_for(rset):
            backup_rset = rset.copy(rset.rows, rset.description)
            with self.subTest(name=self._testname(rset, view.__regid__, "view")):
                self.view(
                    view.__regid__, rset, rset.req.reset_headers(), "main-template"
                )
            # We have to do this because some views modify the
            # resultset's syntax tree
            rset = backup_rset
        for action in self.list_actions_for(rset):
            with self.subTest(name=self._testname(rset, action.__regid__, "action")):
                self._test_action(action)
        for box in self.list_boxes_for(rset):
            self.w_list = []
            w = self.mocked_up_w
            with self.subTest(name=self._testname(rset, box.__regid__, "box")):
                box.render(w)

    def mocked_up_w(self, text, *args, escape=True):
        # we don't care about escape here since we are in a test context
        if not args:
            self.w_list.append(text)
        elif isinstance(args[0], dict):
            self.w_list.append(text % args[0])
        else:
            self.w_list.append(text % args)

    @staticmethod
    def _testname(rset, objid, objtype):
        return "%s_%s_%s" % ("_".join(rset.column_types(0)), objid, objtype)


class AutomaticWebTest(AutoPopulateTest):
    """import this if you wan automatic tests to be ran"""

    tags = AutoPopulateTest.tags | Tags("web", "generated")

    def setUp(self):
        if self.__class__ is AutomaticWebTest:
            # Prevent direct use of AutomaticWebTest to avoid database caching
            # issues.
            return
        super().setUp()

        # access to self.app for proper initialization of the authentication
        # machinery (else some views may fail)
        self.app

    def test_one_each_config(self):
        self.auto_populate(1)
        for rset in self.iter_automatic_rsets(limit=1):
            self._test_everything_for(rset)

    def test_ten_each_config(self):
        self.auto_populate(10)
        for rset in self.iter_automatic_rsets(limit=10):
            self._test_everything_for(rset)

    def test_startup_views(self):
        for vid in self.list_startup_views():
            with self.admin_access.web_request() as req:
                with self.subTest(vid=vid):
                    self.view(vid, None, req)



import os
import os.path as osp
import sys
from subprocess import PIPE, Popen, STDOUT
from unittest import TestCase, main

DATADIR = osp.join(osp.abspath(osp.dirname(__file__)), "data")


def load_po(fname):
    """load a po file and  return a set of encountered (msgid, msgctx)"""
    msgs = set()
    msgid = msgctxt = None
    with open(fname) as fobj:
        for line in fobj:
            if line.strip() in ("", "#"):
                continue
            if line.startswith("msgstr"):
                assert not (msgid, msgctxt) in msgs
                msgs.add((msgid, msgctxt))
                msgid = msgctxt = None
            elif line.startswith("msgid"):
                msgid = line.split(" ", 1)[1][1:-1]
            elif line.startswith("msgctx"):
                msgctxt = line.split(" ", 1)[1][1:-1]
            elif msgid is not None:
                msgid += line[1:-1]
            elif msgctxt is not None:
                msgctxt += line[1:-1]
    return msgs


class CubePotGeneratorTC(TestCase):
    """test case for i18n pot file generator"""

    def test_i18ncube(self):
        env = os.environ.copy()
        if "PYTHONPATH" in env:
            env["PYTHONPATH"] += os.pathsep
        else:
            env["PYTHONPATH"] = ""

        env["PYTHONPATH"] += osp.join(DATADIR, "libpython")

        # launch i18ncube
        cmd = [sys.executable, "-m", "cubicweb", "i18ncube", "i18ntestcube"]
        proc = Popen(cmd, env=env, stdout=PIPE, stderr=STDOUT)
        stdout, _ = proc.communicate()
        msg = stdout.decode(sys.getdefaultencoding(), errors="replace")
        self.assertEqual(proc.returncode, 0, msg=msg)

        # compare en.po.ref and en.po
        cubedir = osp.join(DATADIR, "libpython", "cubicweb_i18ntestcube")
        msgs = load_po(osp.join(cubedir, "i18n", "en.po.ref"))
        newmsgs = load_po(osp.join(cubedir, "i18n", "en.po"))
        self.assertEqual(msgs, newmsgs)


if __name__ == "__main__":
    main()



# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""automatic tests"""
from cubicweb_web.view import AnyRsetView
from cubicweb_web.devtools.testlib import AutoPopulateTest, AutomaticWebTest


class AutomaticWebTest(AutomaticWebTest):
    application_rql = [
        "Any L,F WHERE E is CWUser, E login L, E firstname F",
        "Any L,F,E WHERE E is CWUser, E login L, E firstname F",
        "Any COUNT(X) WHERE X is CWUser",
    ]

    def to_test_etypes(self):
        # We do not really want to test cube views here. So we can drop testing
        # some EntityType. The two Blog types below require the sioc cube that
        # we do not want to add as a dependency.
        etypes = super().to_test_etypes()
        etypes -= {"Blog", "BlogEntry", "CWSession"}
        return etypes


class SomeView(AnyRsetView):
    __regid__ = "someview"

    def call(self):
        self._cw.add_js("spam.js")
        self._cw.add_js("spam.js")


class ManualCubicWebTCs(AutoPopulateTest):
    def test_regr_copy_view(self):
        """regression test: make sure we can ask a copy of a
        composite entity
        """
        with self.admin_access.web_request() as req:
            rset = req.execute('CWUser X WHERE X login "admin"')
            self.view("copy", rset, req=req)

    def test_sortable_js_added(self):
        with self.admin_access.web_request() as req:
            # sortable.js should not be included by default
            rset = req.execute("CWUser X")
            self.assertNotIn(
                b"jquery.tablesorter.js", self.view("oneline", rset, req=req).source
            )

        with self.admin_access.web_request() as req:
            # but should be included by the tableview
            rset = req.execute(
                "Any P,F,S LIMIT 1 WHERE P is CWUser, P firstname F, P surname S"
            )
            self.assertIn(
                b"jquery.tablesorter.js", self.view("table", rset, req=req).source
            )

    def test_js_added_only_once(self):
        with self.admin_access.web_request() as req:
            self.vreg._loadedmods[__name__] = {}
            self.vreg.register(SomeView)
            rset = req.execute("CWUser X")
            source = self.view("someview", rset, req=req).source
            self.assertEqual(source.count(b"spam.js"), 1)

    def test_unrelateddivs(self):
        with self.admin_access.client_cnx() as cnx:
            cnx.create_entity("CWGroup", name="R&D")
            cnx.commit()
        with self.admin_access.web_request(relation="in_group_subject") as req:
            rset = req.execute('Any X WHERE X is CWUser, X login "admin"')
            self.view("unrelateddivs", rset, req=req)


if __name__ == "__main__":
    from logilab.common.testlib import unittest_main

    unittest_main()



from logilab.common.date import datetime2ticks

from cubicweb_web import facet
from cubicweb_web.devtools.testlib import WebCWTC


class BaseFacetCWTC(WebCWTC):
    def prepare_rqlst(
        self,
        req,
        rql="CWUser X",
        mainvar="X",
        expected_baserql="Any X WHERE X is CWUser",
        expected_preparedrql="DISTINCT Any  WHERE X is CWUser",
    ):
        rset = req.cnx.execute(rql)
        rqlst = rset.syntax_tree().copy()
        filtered_variable, baserql = facet.init_facets(
            rset, rqlst.children[0], mainvar=mainvar
        )
        self.assertEqual(filtered_variable.name, mainvar)
        self.assertEqual(baserql, expected_baserql)
        self.assertEqual(rqlst.as_string(), expected_preparedrql)
        return rset, rqlst, filtered_variable

    def _in_group_facet(self, req, cls=facet.RelationFacet, no_relation=False):
        rset, rqlst, filtered_variable = self.prepare_rqlst(req)
        cls.no_relation = no_relation
        f = cls(
            req,
            rset=rset,
            select=rqlst.children[0],
            filtered_variable=filtered_variable,
        )
        f.__regid__ = "in_group"
        f.rtype = "in_group"
        f.role = "subject"
        f.target_attr = "name"
        guests, managers = [
            eid
            for eid, in req.cnx.execute(
                "CWGroup G ORDERBY GN "
                'WHERE G name GN, G name IN ("guests", "managers")'
            )
        ]
        groups = [
            eid
            for eid, in req.cnx.execute(
                "CWGroup G ORDERBY GN "
                'WHERE G name GN, G name IN ("guests", "managers")'
            )
        ]
        return f, groups

    def test_relation_simple(self):
        with self.admin_access.web_request() as req:
            f, (guests, managers) = self._in_group_facet(req)
            self.assertEqual(
                f.vocabulary(), [("guests", guests), ("managers", managers)]
            )
            # ensure rqlst is left unmodified
            self.assertEqual(f.select.as_string(), "DISTINCT Any  WHERE X is CWUser")
            # rqlst = rset.syntax_tree()
            self.assertEqual(f.possible_values(), [str(guests), str(managers)])
            # ensure rqlst is left unmodified
            self.assertEqual(f.select.as_string(), "DISTINCT Any  WHERE X is CWUser")
            f._cw.form[f.__regid__] = str(guests)
            f.add_rql_restrictions()
            # selection is cluttered because rqlst has been prepared for facet (it
            # is not in real life)
            self.assertEqual(
                f.select.as_string(),
                "DISTINCT Any  WHERE X is CWUser, X in_group D, D eid %s" % guests,
            )

    def test_relation_multiple_and(self):
        with self.admin_access.web_request() as req:
            f, (guests, managers) = self._in_group_facet(req)
            f._cw.form[f.__regid__] = [str(guests), str(managers)]
            f._cw.form[f.__regid__ + "_andor"] = "AND"
            f.add_rql_restrictions()
            self.assertEqual(
                f.select.as_string(),
                "DISTINCT Any  WHERE X is CWUser, X in_group A, B eid %s, X in_group B, A eid %s"
                % (guests, managers),
            )

    def test_relation_multiple_or(self):
        with self.admin_access.web_request() as req:
            f, (guests, managers) = self._in_group_facet(req)
            f._cw.form[f.__regid__] = [str(guests), str(managers)]
            f._cw.form[f.__regid__ + "_andor"] = "OR"
            f.add_rql_restrictions()
            self.assertEqual(
                f.select.as_string(),
                "DISTINCT Any  WHERE X is CWUser, X in_group A, A eid IN(%s, %s)"
                % (guests, managers),
            )

    def test_relation_optional_rel(self):
        with self.admin_access.web_request() as req:
            rset = req.cnx.execute(
                "Any X,GROUP_CONCAT(GN) GROUPBY X "
                'WHERE X in_group G?, G name GN, NOT G name "users"'
            )
            rqlst = rset.syntax_tree().copy()
            select = rqlst.children[0]
            filtered_variable, baserql = facet.init_facets(rset, select)

            f = facet.RelationFacet(
                req, rset=rset, select=select, filtered_variable=filtered_variable
            )
            f.rtype = "in_group"
            f.role = "subject"
            f.target_attr = "name"
            guests, managers = [
                eid
                for eid, in req.cnx.execute(
                    "CWGroup G ORDERBY GN "
                    'WHERE G name GN, G name IN ("guests", "managers")'
                )
            ]
            self.assertEqual(
                f.vocabulary(), [("guests", guests), ("managers", managers)]
            )
            # ensure rqlst is left unmodified
            self.assertEqual(
                rqlst.as_string(),
                'DISTINCT Any  WHERE X in_group G?, G name GN, NOT G name "users"',
            )
            # rqlst = rset.syntax_tree()
            self.assertEqual(sorted(f.possible_values()), [str(guests), str(managers)])
            # ensure rqlst is left unmodified
            self.assertEqual(
                rqlst.as_string(),
                'DISTINCT Any  WHERE X in_group G?, G name GN, NOT G name "users"',
            )
            req.form[f.__regid__] = str(guests)
            f.add_rql_restrictions()
            # selection is cluttered because rqlst has been prepared for facet (it
            # is not in real life)
            self.assertEqual(
                f.select.as_string(),
                'DISTINCT Any  WHERE X in_group G?, G name GN, NOT G name "users", X in_group D, D eid %s'
                % guests,
            )

    def test_relation_no_relation_1(self):
        with self.admin_access.web_request() as req:
            f, (guests, managers) = self._in_group_facet(req, no_relation=True)
            self.assertEqual(
                f.vocabulary(), [("guests", guests), ("managers", managers)]
            )
            self.assertEqual(f.possible_values(), [str(guests), str(managers)])
            f._cw.create_entity("CWUser", login="hop", upassword="toto")
            self.assertEqual(
                f.vocabulary(),
                [("<no relation>", ""), ("guests", guests), ("managers", managers)],
            )
            self.assertEqual(f.possible_values(), [str(guests), str(managers), ""])
            f._cw.form[f.__regid__] = ""
            f.add_rql_restrictions()
            self.assertEqual(
                f.select.as_string(),
                "DISTINCT Any  WHERE X is CWUser, NOT X in_group G",
            )

    def test_relation_no_relation_2(self):
        with self.admin_access.web_request() as req:
            f, (guests, managers) = self._in_group_facet(req, no_relation=True)
            f._cw.form[f.__regid__] = ["", guests]
            f.select.save_state()
            f.add_rql_restrictions()
            self.assertEqual(
                f.select.as_string(),
                "DISTINCT Any  WHERE X is CWUser, (NOT X in_group B) OR (X in_group A, A eid %s)"
                % guests,
            )
            f.select.recover()
            self.assertEqual(f.select.as_string(), "DISTINCT Any  WHERE X is CWUser")

    def test_relationattribute(self):
        with self.admin_access.web_request() as req:
            f, (guests, managers) = self._in_group_facet(
                req, cls=facet.RelationAttributeFacet
            )
            self.assertEqual(
                f.vocabulary(), [("guests", "guests"), ("managers", "managers")]
            )
            # ensure rqlst is left unmodified
            self.assertEqual(f.select.as_string(), "DISTINCT Any  WHERE X is CWUser")
            # rqlst = rset.syntax_tree()
            self.assertEqual(f.possible_values(), ["guests", "managers"])
            # ensure rqlst is left unmodified
            self.assertEqual(f.select.as_string(), "DISTINCT Any  WHERE X is CWUser")
            f._cw.form[f.__regid__] = "guests"
            f.add_rql_restrictions()
            # selection is cluttered because rqlst has been prepared for facet (it
            # is not in real life)
            self.assertEqual(
                f.select.as_string(),
                'DISTINCT Any  WHERE X is CWUser, X in_group E, E name "guests"',
            )

    def test_hasrelation(self):
        with self.admin_access.web_request() as req:
            rset, rqlst, filtered_variable = self.prepare_rqlst(req)
            f = facet.HasRelationFacet(
                req,
                rset=rset,
                select=rqlst.children[0],
                filtered_variable=filtered_variable,
            )
            f.__regid__ = "has_group"
            f.rtype = "in_group"
            f.role = "subject"
            f._cw.form[f.__regid__] = "feed me"
            f.add_rql_restrictions()
            self.assertEqual(
                f.select.as_string(),
                "DISTINCT Any  WHERE X is CWUser, EXISTS(X in_group A)",
            )

    def test_daterange(self):
        with self.admin_access.web_request() as req:
            rset, rqlst, filtered_variable = self.prepare_rqlst(req)
            f = facet.DateRangeFacet(
                req,
                rset=rset,
                select=rqlst.children[0],
                filtered_variable=filtered_variable,
            )
            f.rtype = "creation_date"
            mind, maxd = req.cnx.execute(
                "Any MIN(CD), MAX(CD) WHERE X is CWUser, X creation_date CD"
            )[0]
            self.assertEqual(f.vocabulary(), [(str(mind), mind), (str(maxd), maxd)])
            # ensure rqlst is left unmodified
            self.assertEqual(rqlst.as_string(), "DISTINCT Any  WHERE X is CWUser")
            # rqlst = rset.syntax_tree()
            self.assertEqual(f.possible_values(), [str(mind), str(maxd)])
            # ensure rqlst is left unmodified
            self.assertEqual(rqlst.as_string(), "DISTINCT Any  WHERE X is CWUser")
            req.form["%s_inf" % f.__regid__] = str(datetime2ticks(mind))
            req.form["%s_sup" % f.__regid__] = str(datetime2ticks(mind))
            f.add_rql_restrictions()
            # selection is cluttered because rqlst has been prepared for facet (it
            # is not in real life)
            self.assertEqual(
                f.select.as_string(),
                'DISTINCT Any  WHERE X is CWUser, X creation_date >= "%s", '
                'X creation_date <= "%s"'
                % (mind.strftime("%Y/%m/%d"), mind.strftime("%Y/%m/%d")),
            )

    def test_attribute(self):
        with self.admin_access.web_request() as req:
            rset, rqlst, filtered_variable = self.prepare_rqlst(req)
            f = facet.AttributeFacet(
                req,
                rset=rset,
                select=rqlst.children[0],
                filtered_variable=filtered_variable,
            )
            f.rtype = "login"
            self.assertEqual(f.vocabulary(), [("admin", "admin"), ("anon", "anon")])
            # ensure rqlst is left unmodified
            self.assertEqual(rqlst.as_string(), "DISTINCT Any  WHERE X is CWUser")
            # rqlst = rset.syntax_tree()
            self.assertEqual(f.possible_values(), ["admin", "anon"])
            # ensure rqlst is left unmodified
            self.assertEqual(rqlst.as_string(), "DISTINCT Any  WHERE X is CWUser")
            req.form[f.__regid__] = "admin"
            f.add_rql_restrictions()
            # selection is cluttered because rqlst has been prepared for facet (it
            # is not in real life)
            self.assertEqual(
                f.select.as_string(), 'DISTINCT Any  WHERE X is CWUser, X login "admin"'
            )

    def test_bitfield(self):
        with self.admin_access.web_request() as req:
            rset, rqlst, filtered_variable = self.prepare_rqlst(
                req,
                "CWAttribute X WHERE X ordernum XO",
                expected_baserql="Any X WHERE X ordernum XO, X is CWAttribute",
                expected_preparedrql="DISTINCT Any  WHERE X ordernum XO, X is CWAttribute",
            )
            f = facet.BitFieldFacet(
                req,
                rset=rset,
                select=rqlst.children[0],
                filtered_variable=filtered_variable,
            )
            f.choices = [
                (
                    "un",
                    1,
                ),
                (
                    "deux",
                    2,
                ),
            ]
            f.rtype = "ordernum"
            self.assertEqual(f.vocabulary(), [("deux", 2), ("un", 1)])
            # ensure rqlst is left unmodified
            self.assertEqual(
                rqlst.as_string(), "DISTINCT Any  WHERE X ordernum XO, X is CWAttribute"
            )
            # rqlst = rset.syntax_tree()
            self.assertEqual(f.possible_values(), ["2", "1"])
            # ensure rqlst is left unmodified
            self.assertEqual(
                rqlst.as_string(), "DISTINCT Any  WHERE X ordernum XO, X is CWAttribute"
            )
            req.form[f.__regid__] = "3"
            f.add_rql_restrictions()
            # selection is cluttered because rqlst has been prepared for facet (it
            # is not in real life)
            self.assertEqual(
                f.select.as_string(),
                "DISTINCT Any  WHERE X ordernum XO, X is CWAttribute, X ordernum C HAVING 3 = (C & 3)",
            )

    def test_bitfield_0_value(self):
        with self.admin_access.web_request() as req:
            rset, rqlst, filtered_variable = self.prepare_rqlst(
                req,
                "CWAttribute X WHERE X ordernum XO",
                expected_baserql="Any X WHERE X ordernum XO, X is CWAttribute",
                expected_preparedrql="DISTINCT Any  WHERE X ordernum XO, X is CWAttribute",
            )
            f = facet.BitFieldFacet(
                req,
                rset=rset,
                select=rqlst.children[0],
                filtered_variable=filtered_variable,
            )
            f.choices = [
                (
                    "zero",
                    0,
                ),
                (
                    "un",
                    1,
                ),
                (
                    "deux",
                    2,
                ),
            ]
            f.rtype = "ordernum"
            self.assertEqual(f.vocabulary(), [("deux", 2), ("un", 1), ("zero", 0)])
            self.assertEqual(f.possible_values(), ["2", "1", "0"])
            req.form[f.__regid__] = "0"
            f.add_rql_restrictions()
            self.assertEqual(
                f.select.as_string(),
                "DISTINCT Any  WHERE X ordernum XO, X is CWAttribute, X ordernum C HAVING 0 = C",
            )

    def test_rql_path_eid(self):
        with self.admin_access.web_request() as req:
            rset, rqlst, filtered_variable = self.prepare_rqlst(req)

            class RPF(facet.RQLPathFacet):
                path = [("X created_by U"), ("U owned_by O"), ("O login OL")]
                filter_variable = "O"
                label_variable = "OL"

            f = RPF(
                req,
                rset=rset,
                select=rqlst.children[0],
                filtered_variable=filtered_variable,
            )
            self.assertEqual(
                f.vocabulary(),
                [
                    ("admin", req.user.eid),
                ],
            )
            # ensure rqlst is left unmodified
            self.assertEqual(rqlst.as_string(), "DISTINCT Any  WHERE X is CWUser")
            # rqlst = rset.syntax_tree()
            self.assertEqual(
                f.possible_values(),
                [
                    str(req.user.eid),
                ],
            )
            # ensure rqlst is left unmodified
            self.assertEqual(rqlst.as_string(), "DISTINCT Any  WHERE X is CWUser")
            req.form[f.__regid__] = "1"
            f.add_rql_restrictions()
            # selection is cluttered because rqlst has been prepared for facet (it
            # is not in real life)
            self.assertEqual(
                f.select.as_string(),
                "DISTINCT Any  WHERE X is CWUser, X created_by F, F owned_by G, G eid 1",
            )

    def test_rql_path_eid_no_label(self):
        with self.admin_access.web_request() as req:
            rset, rqlst, filtered_variable = self.prepare_rqlst(req)

            class RPF(facet.RQLPathFacet):
                path = [("X created_by U"), ("U owned_by O"), ("O login OL")]
                filter_variable = "O"

            f = RPF(
                req,
                rset=rset,
                select=rqlst.children[0],
                filtered_variable=filtered_variable,
            )
            self.assertEqual(
                f.vocabulary(),
                [
                    (str(req.user.eid), req.user.eid),
                ],
            )

    def test_rql_path_attr(self):
        with self.admin_access.web_request() as req:
            rset, rqlst, filtered_variable = self.prepare_rqlst(req)

            class RPF(facet.RQLPathFacet):
                path = [("X created_by U"), ("U owned_by O"), ("O login OL")]
                filter_variable = "OL"

            f = RPF(
                req,
                rset=rset,
                select=rqlst.children[0],
                filtered_variable=filtered_variable,
            )

            self.assertEqual(
                f.vocabulary(),
                [
                    ("admin", "admin"),
                ],
            )
            # ensure rqlst is left unmodified
            self.assertEqual(rqlst.as_string(), "DISTINCT Any  WHERE X is CWUser")
            self.assertEqual(
                f.possible_values(),
                [
                    "admin",
                ],
            )
            # ensure rqlst is left unmodified
            self.assertEqual(rqlst.as_string(), "DISTINCT Any  WHERE X is CWUser")
            req.form[f.__regid__] = "admin"
            f.add_rql_restrictions()
            # selection is cluttered because rqlst has been prepared for facet (it
            # is not in real life)
            self.assertEqual(
                f.select.as_string(),
                'DISTINCT Any  WHERE X is CWUser, X created_by G, G owned_by H, H login "admin"',
            )

    def test_rql_path_check_filter_label_variable(self):
        with self.admin_access.web_request() as req:
            rset, rqlst, filtered_variable = self.prepareg_aggregat_rqlst(req)

            class RPF(facet.RQLPathFacet):
                path = [("X created_by U"), ("U owned_by O"), ("O login OL")]
                filter_variable = "OL"
                label_variable = "OL"

            self.assertRaises(
                AssertionError,
                RPF,
                req,
                rset=rset,
                select=rqlst.children[0],
                filtered_variable=filtered_variable,
            )

    def test_rqlpath_range(self):
        with self.admin_access.web_request() as req:
            rset, rqlst, filtered_variable = self.prepare_rqlst(req)

            class RRF(facet.DateRangeRQLPathFacet):
                path = [("X created_by U"), ("U owned_by O"), ("O creation_date OL")]
                filter_variable = "OL"

            f = RRF(
                req,
                rset=rset,
                select=rqlst.children[0],
                filtered_variable=filtered_variable,
            )
            mind, maxd = req.cnx.execute(
                "Any MIN(CD), MAX(CD) WHERE X is CWUser, X created_by U, U owned_by O, O creation_date CD"
            )[0]
            self.assertEqual(f.vocabulary(), [(str(mind), mind), (str(maxd), maxd)])
            # ensure rqlst is left unmodified
            self.assertEqual(rqlst.as_string(), "DISTINCT Any  WHERE X is CWUser")
            self.assertEqual(f.possible_values(), [str(mind), str(maxd)])
            # ensure rqlst is left unmodified
            self.assertEqual(rqlst.as_string(), "DISTINCT Any  WHERE X is CWUser")
            req.form["%s_inf" % f.__regid__] = str(datetime2ticks(mind))
            req.form["%s_sup" % f.__regid__] = str(datetime2ticks(mind))
            f.add_rql_restrictions()
            # selection is cluttered because rqlst has been prepared for facet (it
            # is not in real life)
            self.assertEqual(
                f.select.as_string(),
                'DISTINCT Any  WHERE X is CWUser, X created_by G, G owned_by H, H creation_date >= "%s", '
                'H creation_date <= "%s"'
                % (mind.strftime("%Y/%m/%d"), mind.strftime("%Y/%m/%d")),
            )

    def prepareg_aggregat_rqlst(self, req):
        return self.prepare_rqlst(
            req,
            "Any 1, COUNT(X) WHERE X is CWUser, X creation_date XD, "
            "X modification_date XM, Y creation_date YD, Y is CWGroup "
            "HAVING DAY(XD)>=DAY(YD) AND DAY(XM)<=DAY(YD)",
            "X",
            expected_baserql="Any 1,COUNT(X) WHERE X is CWUser, X creation_date XD, "
            "X modification_date XM, Y creation_date YD, Y is CWGroup "
            "HAVING DAY(XD) >= DAY(YD), DAY(XM) <= DAY(YD)",
            expected_preparedrql="DISTINCT Any  WHERE X is CWUser, X creation_date XD, "
            "X modification_date XM, Y creation_date YD, Y is CWGroup "
            "HAVING DAY(XD) >= DAY(YD), DAY(XM) <= DAY(YD)",
        )

    def test_aggregat_query_cleanup_select(self):
        with self.admin_access.web_request() as req:
            rset, rqlst, filtered_variable = self.prepareg_aggregat_rqlst(req)
            select = rqlst.children[0]
            facet.cleanup_select(select, filtered_variable=filtered_variable)
            self.assertEqual(
                select.as_string(),
                "DISTINCT Any  WHERE X is CWUser, X creation_date XD, "
                "X modification_date XM, Y creation_date YD, Y is CWGroup "
                "HAVING DAY(XD) >= DAY(YD), DAY(XM) <= DAY(YD)",
            )

    def test_aggregat_query_rql_path(self):
        with self.admin_access.web_request() as req:
            rset, rqlst, filtered_variable = self.prepareg_aggregat_rqlst(req)

            class RPF(facet.RQLPathFacet):
                path = [("X created_by U"), ("U owned_by O"), ("O login OL")]
                filter_variable = "OL"

            f = RPF(
                req,
                rset=rset,
                select=rqlst.children[0],
                filtered_variable=filtered_variable,
            )
            self.assertEqual(f.vocabulary(), [("admin", "admin")])
            self.assertEqual(f.possible_values(), ["admin"])
            req.form[f.__regid__] = "admin"
            f.add_rql_restrictions()
            self.assertEqual(
                f.select.as_string(),
                "DISTINCT Any  WHERE X is CWUser, X creation_date XD, "
                "X modification_date XM, Y creation_date YD, Y is CWGroup, "
                'X created_by G, G owned_by H, H login "admin" '
                "HAVING DAY(XD) >= DAY(YD), DAY(XM) <= DAY(YD)",
            )

    def test_aggregat_query_attribute(self):
        with self.admin_access.web_request() as req:
            rset, rqlst, filtered_variable = self.prepareg_aggregat_rqlst(req)
            f = facet.AttributeFacet(
                req,
                rset=rset,
                select=rqlst.children[0],
                filtered_variable=filtered_variable,
            )
            f.rtype = "login"
            self.assertEqual(f.vocabulary(), [("admin", "admin"), ("anon", "anon")])
            self.assertEqual(f.possible_values(), ["admin", "anon"])
            req.form[f.__regid__] = "admin"
            f.add_rql_restrictions()
            self.assertEqual(
                f.select.as_string(),
                "DISTINCT Any  WHERE X is CWUser, X creation_date XD, "
                'X modification_date XM, Y creation_date YD, Y is CWGroup, X login "admin" '
                "HAVING DAY(XD) >= DAY(YD), DAY(XM) <= DAY(YD)",
            )


if __name__ == "__main__":
    from logilab.common.testlib import unittest_main

    unittest_main()



# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from logilab.common.testlib import unittest_main
from logilab.mtconverter import html_unescape

from cubicweb.devtools import BASE_URL
from cubicweb.utils import json

from cubicweb_web.devtools.testlib import WebCWTC
from cubicweb_web.view import StartupView, TRANSITIONAL_DOCTYPE
from cubicweb_web.views import vid_from_rset


def loadjson(value):
    return json.loads(html_unescape(value))


class VidFromRsetCWTC(WebCWTC):
    def test_no_rset(self):
        with self.admin_access.web_request() as req:
            self.assertEqual(vid_from_rset(req, None, self.schema), "index")

    def test_no_entity(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('Any X WHERE X login "blabla"')
            self.assertEqual(vid_from_rset(req, rset, self.schema), "noresult")

    def test_one_entity(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('Any X WHERE X login "admin"')
            self.assertEqual(vid_from_rset(req, rset, self.schema), "primary")
            rset = req.execute('Any X, L WHERE X login "admin", X login L')
            self.assertEqual(vid_from_rset(req, rset, self.schema), "primary")
            req.search_state = ("pasnormal",)
            rset = req.execute('Any X WHERE X login "admin"')
            self.assertEqual(
                vid_from_rset(req, rset, self.schema), "outofcontext-search"
            )

    def test_one_entity_eid(self):
        with self.admin_access.web_request() as req:
            rset = req.execute("Any X WHERE X eid 1")
            self.assertEqual(vid_from_rset(req, rset, self.schema), "primary")

    def test_more_than_one_entity_same_type(self):
        with self.admin_access.web_request() as req:
            rset = req.execute("Any X WHERE X is CWUser")
            self.assertEqual(vid_from_rset(req, rset, self.schema), "sameetypelist")
            rset = req.execute("Any X, L WHERE X login L")
            self.assertEqual(vid_from_rset(req, rset, self.schema), "sameetypelist")

    def test_more_than_one_entity_diff_type(self):
        with self.admin_access.web_request() as req:
            rset = req.execute("Any X WHERE X is IN (CWUser, CWGroup)")
            self.assertEqual(vid_from_rset(req, rset, self.schema), "list")

    def test_more_than_one_entity_by_row(self):
        with self.admin_access.web_request() as req:
            rset = req.execute("Any X, G WHERE X in_group G")
            self.assertEqual(vid_from_rset(req, rset, self.schema), "table")

    def test_more_than_one_entity_by_row_2(self):
        with self.admin_access.web_request() as req:
            rset = req.execute("Any X, GN WHERE X in_group G, G name GN")
            self.assertEqual(vid_from_rset(req, rset, self.schema), "table")

    def test_aggregat(self):
        with self.admin_access.web_request() as req:
            rset = req.execute("Any X, COUNT(T) GROUPBY X WHERE X is T")
            self.assertEqual(vid_from_rset(req, rset, self.schema), "table")
            rset = req.execute("Any MAX(X) WHERE X is CWUser")
            self.assertEqual(vid_from_rset(req, rset, self.schema), "table")

    def test_subquery(self):
        with self.admin_access.web_request() as req:
            rset = req.execute(
                "DISTINCT Any X,N ORDERBY N "
                "WITH X,N BEING ("
                "     (DISTINCT Any P,N WHERE P is CWUser, P login N)"
                "       UNION"
                "     (DISTINCT Any W,N WHERE W is CWGroup, W name N))"
            )
            self.assertEqual(vid_from_rset(req, rset, self.schema), "table")


class TableViewCWTC(WebCWTC):
    def _prepare_entity(self, req):
        e = req.create_entity("State", name="<toto>", description='loo"ong blabla')
        rset = req.execute(
            "Any X, D, CD, NOW - CD WHERE X is State, "
            "X description D, X creation_date CD, X eid %(x)s",
            {"x": e.eid},
        )
        view = self.vreg["views"].select("table", req, rset=rset)
        return e, rset, view

    def test_sortvalue(self):
        with self.admin_access.web_request() as req:
            e, _, view = self._prepare_entity(req)
            colrenderers = view.build_column_renderers()[:3]
            self.assertListEqual(
                [renderer.sortvalue(0) for renderer in colrenderers],
                ["<toto>", 'loo"ong blabla', e.creation_date],
            )


class HTMLStreamTests(WebCWTC):
    def test_set_doctype_reset_xmldecl(self):
        """
        tests `cubicweb.web.request.CubicWebRequestBase.set_doctype`
        with xmldecl reset
        """

        class MyView(StartupView):
            __regid__ = "my-view"

            def call(self):
                self._cw.set_doctype("<!DOCTYPE html>")

        with self.admin_access.web_request() as req:
            with self.temporary_appobjects(MyView):
                html_source = self.view("my-view", req=req).source
                source_lines = [
                    line.strip()
                    for line in html_source.splitlines(False)
                    if line.strip()
                ]
                self.assertListEqual(
                    [
                        b"<!DOCTYPE html>",
                        b'<html xmlns:cubicweb="http://www.cubicweb.org" lang="en">',
                    ],
                    source_lines[:2],
                )

    def test_set_doctype_no_reset_xmldecl(self):
        """
        tests `cubicweb.web.request.CubicWebRequestBase.set_doctype`
        with no xmldecl reset
        """
        html_doctype = TRANSITIONAL_DOCTYPE.strip()

        class MyView(StartupView):
            __regid__ = "my-view"

            def call(self):
                self._cw.set_doctype(html_doctype)
                self._cw.main_stream.set_htmlattrs([("lang", "cz")])

        with self.admin_access.web_request() as req:
            with self.temporary_appobjects(MyView):
                html_source = self.view("my-view", req=req).source
                source_lines = [
                    line.strip()
                    for line in html_source.splitlines(False)
                    if line.strip()
                ]
                self.assertListEqual(
                    [
                        html_doctype.encode("ascii"),
                        b'<html xmlns:cubicweb="http://www.cubicweb.org" lang="cz">',
                        b"<head>",
                    ],
                    source_lines[:3],
                )


class BaseViewsCWTC(WebCWTC):
    def test_null(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('Any X WHERE X login "admin"')
            result = req.view("null", rset)
            self.assertEqual(result, "")

    def test_final(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('Any "<script></script>"')
            result = req.view("final", rset)
            self.assertEqual(result, "&lt;script&gt;&lt;/script&gt;")

    def test_incontext(self):
        with self.admin_access.web_request() as req:
            entity = req.create_entity(
                "CWUser", login="<script></script>", upassword="toto"
            )
            result = entity.view("incontext")
            expected = (
                f'<a href="{BASE_URL}{entity.eid}" title="">'
                f"&lt;script&gt;&lt;/script&gt;</a>"
            )
            self.assertEqual(result, expected)

    def test_outofcontext(self):
        with self.admin_access.web_request() as req:
            entity = req.create_entity(
                "CWUser", login="<script></script>", upassword="toto"
            )
            result = entity.view("outofcontext")
            expect = (
                f'<a href="{BASE_URL}{entity.eid}" title="">'
                f"&lt;script&gt;&lt;/script&gt;</a>"
            )
            self.assertEqual(result, expect)

    def test_outofcontext_2(self):
        with self.admin_access.web_request() as req:
            entity = req.create_entity(
                "CWUser", login="<script></script>", upassword="toto"
            )
            result = entity.view("oneline")
            expect = (
                f'<a href="{BASE_URL}{entity.eid}" title="">'
                f"&lt;script&gt;&lt;/script&gt;</a>"
            )
            self.assertEqual(result, expect)

    def test_text(self):
        with self.admin_access.web_request() as req:
            entity = req.create_entity(
                "CWUser", login="<script></script>", upassword="toto"
            )
            result = entity.view("text")
            self.assertEqual(result, "<script></script>")

    def test_textincontext(self):
        with self.admin_access.web_request() as req:
            entity = req.create_entity(
                "CWUser", login="<script></script>", upassword="toto"
            )
            result = entity.view("textincontext")
            self.assertEqual(result, "<script></script>")

    def test_textoutofcontext(self):
        with self.admin_access.web_request() as req:
            entity = req.create_entity(
                "CWUser", login="<script></script>", upassword="toto"
            )
            result = entity.view("textoutofcontext")
            self.assertEqual(result, "<script></script>")

    def test_list(self):
        with self.admin_access.web_request() as req:
            entity = req.create_entity(
                "CWUser", login="<script></script>", upassword="toto"
            )
            rset = req.execute("Any X WHERE X is CWUser")
            result = req.view("list", rset)
            expected = f"""<ul class="section">
<li><a href="{BASE_URL}{entity.eid}" title="">&lt;script&gt;&lt;/script&gt;</a></li>
<li><a href="{BASE_URL}cwuser/admin" title="">admin</a></li>
<li><a href="{BASE_URL}cwuser/anon" title="">anon</a></li>
</ul>
"""
            self.assertEqual(result, expected)

    def test_simplelist(self):
        with self.admin_access.web_request() as req:
            entity = req.create_entity(
                "CWUser", login="<script></script>", upassword="toto"
            )
            rset = req.execute("Any X WHERE X is CWUser")
            result = req.view("simplelist", rset)
            expected = (
                f'<div class="section">'
                f'<a href="{BASE_URL}{entity.eid}" title="">'
                f"&lt;script&gt;&lt;/script&gt;</a></div>"
                f'<div class="section">'
                f'<a href="{BASE_URL}cwuser/admin" title="">admin</a></div>'
                f'<div class="section">'
                f'<a href="{BASE_URL}cwuser/anon" title="">anon</a></div>'
            )
            self.assertEqual(result, expected)

    def test_sameetypelist(self):
        with self.admin_access.web_request() as req:
            entity = req.create_entity(
                "CWUser", login="<script></script>", upassword="toto"
            )
            rset = req.execute("Any X WHERE X is CWUser")
            result = req.view("sameetypelist", rset)
            expected = (
                f"<h1>CWUser_plural</h1>"
                f'<div class="section">'
                f'<a href="{BASE_URL}{entity.eid}" title="">'
                f"&lt;script&gt;&lt;/script&gt;</a></div>"
                f'<div class="section">'
                f'<a href="{BASE_URL}cwuser/admin" title="">admin</a></div>'
                f'<div class="section">'
                f'<a href="{BASE_URL}cwuser/anon" title="">anon</a></div>'
            )
            self.assertEqual(expected, result)

    def test_sameetypelist2(self):
        with self.admin_access.web_request() as req:
            entity = req.create_entity(
                "CWUser", login="<script></script>", upassword="toto"
            )
            rset = req.execute("Any X WHERE X is CWUser")
            result = req.view("csv", rset)
            expected = (
                f'<a href="{BASE_URL}{entity.eid}" title="">'
                f"&lt;script&gt;&lt;/script&gt;</a>, "
                f'<a href="{BASE_URL}cwuser/admin" title="">admin</a>, '
                f'<a href="{BASE_URL}cwuser/anon" title="">anon</a>'
            )
            self.assertEqual(result, expected)

    def test_metadata(self):
        with self.admin_access.web_request() as req:
            entity = req.create_entity(
                "CWUser", login="<script></script>", upassword="toto"
            )
            entity.cw_set(creation_date="2000-01-01 00:00:00")
            entity.cw_set(modification_date="2015-01-01 00:00:00")
            result = entity.view("metadata")
            expected = (
                "<div>CWUser #%d - <span>latest update on</span>"
                ' <span class="value">2015/01/01</span>,'
                " <span>created on</span>"
                ' <span class="value">2000/01/01</span></div>' % entity.eid
            )
            self.assertEqual(result, expected)


if __name__ == "__main__":
    unittest_main()



# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""web ui configuration for cubicweb instances"""


import hmac
import os
from os.path import dirname, join, exists, split, isdir
from uuid import uuid4

from logilab.common.configuration import Method, merge_options
from logilab.common.decorators import cached, cachedproperty

from cubicweb import _
from cubicweb.cwconfig import CubicWebConfiguration, register_persistent_options
from cubicweb.pyramid.config import AllInOneConfiguration

_DATA_DIR = join(dirname(__file__), "data")


register_persistent_options(
    (
        # site-wide only web ui configuration
        (
            "site-title",
            {
                "type": "string",
                "default": "unset title",
                "help": _("site title"),
                "sitewide": True,
                "group": "ui",
            },
        ),
        (
            "main-template",
            {
                "type": "string",
                "default": "main-template",
                "help": _("id of main template used to render pages"),
                "sitewide": True,
                "group": "ui",
            },
        ),
        # user web ui configuration
        (
            "fckeditor",
            {
                "type": "yn",
                "default": False,
                "help": _(
                    "should html fields being edited using fckeditor (a HTML "
                    "WYSIWYG editor).  You should also select text/html as default "
                    "text format to actually get fckeditor."
                ),
                "group": "ui",
            },
        ),
        # navigation configuration
        (
            "page-size",
            {
                "type": "int",
                "default": 40,
                "help": _("maximum number of objects displayed by page of results"),
                "group": "navigation",
            },
        ),
        (
            "related-limit",
            {
                "type": "int",
                "default": 8,
                "help": _(
                    "maximum number of related entities to display in the primary "
                    "view"
                ),
                "group": "navigation",
            },
        ),
        (
            "combobox-limit",
            {
                "type": "int",
                "default": 20,
                "help": _("maximum number of entities to display in related combo box"),
                "group": "navigation",
            },
        ),
    )
)


class WebConfiguration(CubicWebConfiguration):
    """web instance (in a web server) client of a RQL server"""

    cube_appobject_path = CubicWebConfiguration.cube_appobject_path | {"views"}

    options = merge_options(
        CubicWebConfiguration.options
        + (
            (
                "use-uicache",
                {
                    "type": "yn",
                    "default": True,
                    "help": _("should css be compiled and store in uicache"),
                    "group": "ui",
                    "level": 2,
                },
            ),
            (
                "query-log-file",
                {
                    "type": "string",
                    "default": None,
                    "help": "web instance query log file",
                    "group": "web",
                    "level": 3,
                },
            ),
            # web configuration
            (
                "datadir-url",
                {
                    "type": "string",
                    "default": None,
                    "help": (
                        'base url for static data, if different from "${base-url}/data/".  '
                        "If served from a different domain, that domain should allow "
                        "cross-origin requests."
                    ),
                    "group": "web",
                },
            ),
            (
                "auth-mode",
                {
                    "type": "choice",
                    "choices": ("cookie", "http"),
                    "default": "cookie",
                    "help": "authentication mode (cookie / http)",
                    "group": "web",
                    "level": 3,
                },
            ),
            (
                "realm",
                {
                    "type": "string",
                    "default": "cubicweb",
                    "help": "realm to use on HTTP authentication mode",
                    "group": "web",
                    "level": 3,
                },
            ),
            (
                "http-session-time",
                {
                    "type": "time",
                    "default": 0,
                    "help": "duration of the cookie used to store session identifier. "
                    "If 0, the cookie will expire when the user exist its browser. "
                    "Should be 0 or greater than repository's session-time.",
                    "group": "web",
                    "level": 2,
                },
            ),
            (
                "submit-mail",
                {
                    "type": "string",
                    "default": None,
                    "help": (
                        "Mail used as recipient to report bug in this instance, "
                        "if you want this feature on"
                    ),
                    "group": "web",
                    "level": 2,
                },
            ),
            (
                "language-mode",
                {
                    "type": "choice",
                    "choices": ("http-negotiation", "url-prefix", ""),
                    "default": "http-negotiation",
                    "help": (
                        "source for interface's language detection. "
                        'If set to "http-negotiation" the Accept-Language HTTP header will be used,'
                        ' if set to "url-prefix", the URL will be inspected for a'
                        " short language prefix."
                    ),
                    "group": "web",
                    "level": 2,
                },
            ),
            (
                "print-traceback",
                {
                    "type": "yn",
                    "default": CubicWebConfiguration.mode != "system",
                    "help": "print the traceback on the error page when an error occurred",
                    "group": "web",
                    "level": 2,
                },
            ),
            (
                "captcha-font-file",
                {
                    "type": "string",
                    "default": join(_DATA_DIR, "porkys.ttf"),
                    "help": (
                        "True type font to use for captcha image generation (you must have the "
                        "python imaging library installed to use captcha)"
                    ),
                    "group": "web",
                    "level": 3,
                },
            ),
            (
                "captcha-font-size",
                {
                    "type": "int",
                    "default": 25,
                    "help": (
                        "Font size to use for captcha image generation (you must have the python "
                        "imaging library installed to use captcha)"
                    ),
                    "group": "web",
                    "level": 3,
                },
            ),
            (
                "concat-resources",
                {
                    "type": "yn",
                    "default": False,
                    "help": "use modconcat-like URLS to concat and serve JS / CSS files",
                    "group": "web",
                    "level": 2,
                },
            ),
            (
                "anonymize-jsonp-queries",
                {
                    "type": "yn",
                    "default": True,
                    "help": "anonymize the connection before executing any jsonp query.",
                    "group": "web",
                    "level": 1,
                },
            ),
            (
                "generate-staticdir",
                {
                    "type": "yn",
                    "default": False,
                    "help": "Generate the static data resource directory on upgrade.",
                    "group": "web",
                    "level": 2,
                },
            ),
            (
                "staticdir-path",
                {
                    "type": "string",
                    "default": None,
                    "help": "The static data resource directory path.",
                    "group": "web",
                    "level": 2,
                },
            ),
            # ctl configuration
            (
                "max-post-length",  # XXX specific to "wsgi" server
                {
                    "type": "bytes",
                    "default": "100MB",
                    "help": "maximum length of HTTP request. Default to 100 MB.",
                    "group": "web",
                    "level": 1,
                },
            ),
            (
                "pid-file",
                {
                    "type": "string",
                    "default": Method("default_pid_file"),
                    "help": "repository's pid file",
                    "group": "main",
                    "level": 2,
                },
            ),
        )
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.uiprops = None
        self.datadir_url = None
        self._debug = None
        # don't register ReStructured Text directives by simple import, avoid pb
        # with eg sphinx.
        # XXX should be done properly with a function from cw.uicfg
        try:
            from cubicweb_web.ext.rest import cw_rest_init
        except ImportError:
            pass
        else:
            cw_rest_init()

    @property
    def in_debug_mode(self):
        """
        we need all of this because self["debug"] is regularely overwritten in
        the workflow because the configuration is not initialy loaded and
        loaded after most includeme.
        """
        if self._debug is not None:
            return self._debug

        return self["debug"]

    @in_debug_mode.setter
    def in_debug_mode(self, value):
        self._debug = value

    def fckeditor_installed(self):
        if self.uiprops is None:
            return False
        return exists(self.uiprops.get("FCKEDITOR_PATH", ""))

    def cwproperty_definitions(self):
        for key, pdef in super().cwproperty_definitions():
            if key == "ui.fckeditor" and not self.fckeditor_installed():
                continue
            yield key, pdef

    @cachedproperty
    def _instance_salt(self):
        """This random key/salt is used to sign content to be sent back by
        browsers, eg. in the error report form.
        """
        return str(uuid4()).encode("ascii")

    def sign_text(self, text):
        """sign some text for later checking"""
        # hmac.new expect bytes
        if isinstance(text, str):
            text = text.encode("utf-8")
        # replace \r\n so we do not depend on whether a browser "reencode"
        # original message using \r\n or not
        return hmac.new(
            self._instance_salt,
            text.strip().replace(b"\r\n", b"\n"),
            digestmod="sha3_512",
        ).hexdigest()

    def check_text_sign(self, text, signature):
        """check the text signature is equal to the given signature"""
        return self.sign_text(text) == signature

    def locate_resource(self, rid):
        """return the (directory, filename) where the given resource
        may be found
        """
        return self._fs_locate(rid, "data")

    def locate_doc_file(self, fname):
        """return the directory where the given resource may be found"""
        return self._fs_locate(fname, "wdoc")[0]

    @cached
    def _fs_path_locate(self, rid, rdirectory):
        """return the directory where the given resource may be found"""
        path = [self.apphome] + self.cubes_path() + [dirname(__file__)]
        for directory in path:
            if exists(join(directory, rdirectory, rid)):
                return directory

    def _fs_locate(self, rid, rdirectory):
        """return the (directory, filename) where the given resource
        may be found
        """
        directory = self._fs_path_locate(rid, rdirectory)
        if directory is None:
            return None, None
        if self["use-uicache"] and rdirectory == "data" and rid.endswith(".css"):
            return (
                self.ensure_uid_directory(
                    self.uiprops.process_resource(join(directory, rdirectory), rid)
                ),
                rid,
            )
        return join(directory, rdirectory), rid

    def locate_all_files(self, rid, rdirectory="wdoc"):
        """return all files corresponding to the given resource"""
        path = [self.apphome] + self.cubes_path() + [dirname(__file__)]
        for directory in path:
            fpath = join(directory, rdirectory, rid)
            if exists(fpath):
                yield join(fpath)

    def load_configuration(self, **kw):
        """load instance's configuration files"""
        super().load_configuration(**kw)
        # load external resources definition
        self._build_ui_properties()

    def _init_base_url(self):
        super()._init_base_url()

        self.datadir_url = self["datadir-url"]
        if self.datadir_url:
            if self.datadir_url[-1] != "/":
                self.datadir_url += "/"
            if self.mode != "test":
                self.datadir_url += "%s/" % self.instance_md5_version()
            return

        data_relpath = self.data_relpath()
        self.datadir_url = self._generate_base_url() + data_relpath

    def data_relpath(self):
        if self.mode == "test":
            return "data/"
        return "data/%s/" % self.instance_md5_version()

    def _build_ui_properties(self):
        # self.datadir_url[:-1] to remove trailing /
        from cubicweb_web.propertysheet import PropertySheet

        cachedir = join(self.appdatahome, "uicache")
        self.check_writeable_uid_directory(cachedir)
        self.uiprops = PropertySheet(
            cachedir,
            data=lambda x: self.datadir_url + x,
            datadir_url=self.datadir_url[:-1],
        )
        self._init_uiprops(self.uiprops)

    def _init_uiprops(self, uiprops):
        libuiprops = join(_DATA_DIR, "uiprops.py")
        uiprops.load(libuiprops)
        for path in reversed([self.apphome] + self.cubes_path()):
            self._load_ui_properties_file(uiprops, path)
        self._load_ui_properties_file(uiprops, self.apphome)
        datadir_url = uiprops.context["datadir_url"]
        cubicweb_js_url = datadir_url + "/cubicweb.js"
        if cubicweb_js_url not in uiprops["JAVASCRIPTS"]:
            uiprops["JAVASCRIPTS"].insert(0, cubicweb_js_url)

    def _load_ui_properties_file(self, uiprops, path):
        uipropsfile = join(path, "uiprops.py")
        if exists(uipropsfile):
            self.debug("loading %s", uipropsfile)
            uiprops.load(uipropsfile)

    # static files handling ###################################################

    @property
    def static_directory(self):
        return join(self.appdatahome, "static")

    def static_file_exists(self, rpath):
        return exists(join(self.static_directory, rpath))

    def static_file_open(self, rpath, mode="wb"):
        staticdir = self.static_directory
        rdir, filename = split(rpath)
        if rdir:
            staticdir = join(staticdir, rdir)
            if not isdir(staticdir) and "w" in mode:
                self.check_writeable_uid_directory(staticdir)
        return open(join(staticdir, filename), mode)

    def static_file_add(self, rpath, data):
        stream = self.static_file_open(rpath)
        stream.write(data)
        stream.close()
        self.ensure_uid(rpath)

    def static_file_del(self, rpath):
        if self.static_file_exists(rpath):
            os.remove(join(self.static_directory, rpath))


class WebAllInOneConfiguration(AllInOneConfiguration, WebConfiguration):
    """web instance (in a web server) client of a RQL server"""

    name = "all-in-one"
    options = merge_options(WebConfiguration.options + AllInOneConfiguration.options)

    cubicweb_appobject_path = (
        WebConfiguration.cubicweb_appobject_path
        | AllInOneConfiguration.cubicweb_appobject_path
    )
    cube_appobject_path = (
        WebConfiguration.cube_appobject_path | AllInOneConfiguration.cube_appobject_path
    )



# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""xbel views"""


from cubicweb import _

from cubicweb.predicates import is_instance
from cubicweb_web.view import EntityView
from cubicweb_web.views.xmlrss import XMLView


class XbelView(XMLView):
    __regid__ = "xbel"
    title = _("xbel export")
    templatable = False
    content_type = "text/xml"  # application/xbel+xml

    def cell_call(self, row, col):
        self.wview("xbelitem", self.cw_rset, row=row, col=col)

    def call(self):
        """display a list of entities by calling their <item_vid> view"""
        self.w('<?xml version="1.0" encoding="%s"?>\n', self._cw.encoding)
        self.w(
            '<!DOCTYPE xbel PUBLIC "+//IDN python.org//DTD XML Bookmark Exchange Language 1.0//EN//XML" "http://www.python.org/topics/xml/dtds/xbel-1.0.dtd">'
        )
        self.w('<xbel version="1.0">')
        self.w("<title>%s</title>", self._cw._("bookmarks"))
        for i in range(self.cw_rset.rowcount):
            self.cell_call(i, 0)
        self.w("</xbel>")


class XbelItemView(EntityView):
    __regid__ = "xbelitem"

    def cell_call(self, row, col):
        entity = self.cw_rset.complete_entity(row, col)
        self.w('<bookmark href="%s">', self.url(entity))
        self.w("  <title>%s</title>", entity.dc_title())
        self.w("</bookmark>")

    def url(self, entity):
        return entity.absolute_url()


class XbelItemBookmarkView(XbelItemView):
    __select__ = is_instance("Bookmark")

    def url(self, entity):
        return entity.actual_url()
